from __future__ import annotations

from asyncio import Protocol
from collections import defaultdict, deque
from collections.abc import Callable

import msgspec
import msgspec.json
import msgspec.msgpack

from agentica_internal.core.print import colorize, hdiv, tprint
from agentica_internal.warpc.messages import DefinitionMsg, ListMsg, RPCMsg, StrMsg, VarsMsg
from agentica_internal.warpc.msg.term import TermMsg
from agentica_internal.warpc.worlds.interface import AsyncRecvBytes, AsyncSendBytes
from agentica_internal.warpc_transcode.conv_utils import (
    LOG_TS,
    ORDER_MARKER,
    MangleManager,
    msg_TS_name_restoration,
    msg_uni_name_sanitization,
    order_defs,
)
from agentica_internal.warpc_transcode.py_to_uni import py_to_uni_rpc
from agentica_internal.warpc_transcode.uni_msgs import (
    DefnUID,
    DefUniMsg,
    FutureCompletedUniMsg,
    RequestUniMsg,
    ResUniMsg,
    RpcUniMsg,
    TermUnion,
    UniMsg,
    json_to_concept_uni,
    json_to_rpc_uni,
    uni_to_json,
)
from agentica_internal.warpc_transcode.uni_to_py import uni_to_py_def, uni_to_py_rpc, uni_to_py_term


class InterceptorProto(Protocol):
    def decode_defs(self, warpc_locals: bytes) -> bytes: ...
    def intercept_sdk(
        self, recv_from_sdk: AsyncRecvBytes, send_to_sdk: AsyncSendBytes
    ) -> tuple[AsyncRecvBytes, AsyncSendBytes]: ...


class NoopInterceptor(InterceptorProto):
    def decode_defs(self, warpc_locals: bytes) -> bytes:
        return warpc_locals

    def intercept_sdk(
        self, recv_from_sdk: AsyncRecvBytes, send_to_sdk: AsyncSendBytes
    ) -> tuple[AsyncRecvBytes, AsyncSendBytes]:
        return recv_from_sdk, send_to_sdk


TRANS = colorize('Transcoder')


class TranscodingInterceptor(InterceptorProto):
    defn_ctx: dict[DefnUID, DefUniMsg]
    mangler: MangleManager
    _orig_recv_bytes: AsyncRecvBytes | None
    _orig_send_bytes: AsyncSendBytes | None
    logging: bool

    def __init__(
        self,
        log: Callable[..., None] | None = None,
    ) -> None:
        self._msg_counter = 0
        self._in_queue: deque[RPCMsg] = deque()
        self.defn_ctx = {}
        self.mangler = MangleManager()
        self.logging = bool(LOG_TS)
        self.log("init")

    def log(self, *args) -> None:
        # note to others: you can either use FORE_ENABLE_LOGGING=TS, or
        # ./run_local.sh --tags=TS ...
        # ./run_local_pylog.sh test_something TS
        # or even just setting self.logging = True temporarily
        if self.logging:
            tprint(TRANS, *args)

    def log_msg(self, msg: UniMsg | TermMsg):
        if self.logging:
            msg.pprint()
            hdiv()

    def decode_defs(self, warpc_locals: bytes) -> bytes:
        self.log("decode_defs")

        # Case of empty warpc_globals
        if not warpc_locals or not warpc_locals.strip():
            return msgspec.msgpack.Encoder().encode({})
        dict_decoder = msgspec.json.Decoder(dict[str, msgspec.Raw])
        transcoded_vars: dict[str, TermMsg] = {}
        def_msgs: list['DefinitionMsg'] = []

        raw_dict = dict_decoder.decode(warpc_locals)

        # Populate context and map defs uid to set of names of locals with that defn
        uid_to_local_name: defaultdict[DefnUID, set[str]] = defaultdict(set)
        locals_without_uid: dict[str, TermUnion] = {}
        for name, json_bytes in raw_dict.items():
            if name == "":
                raise ValueError("Var cannot be empty")

            try:
                term_msg = json_to_concept_uni(json_bytes)
                # Sanitize names from TypeScript by mangling TS identifiers to Python identifiers
                term_msg = msg_uni_name_sanitization(term_msg, self.mangler)
                name = self.mangler.mangle(name)
            except Exception as e:
                self.log("error json bytes; json_bytes =")
                # print("BYTES", bytes(json_bytes))
                raise e

            self.log("locals parsing incoming, var ", name, "| parsed:")
            self.log_msg(term_msg)

            if hasattr(term_msg, 'uid') and term_msg.uid is not None:
                assert isinstance(term_msg, DefUniMsg), f"Locals with uid should be definitions"
                self.defn_ctx[term_msg.uid] = term_msg
                uid_to_local_name[term_msg.uid].add(name)
            else:
                locals_without_uid[name] = term_msg

        # Order defs by referential dependencies
        ordered_defs, _ = order_defs(list(self.defn_ctx.values()))

        self.log(f"@@@ uid_to_local_name: {uid_to_local_name}")

        # Transcode definitions and collect Vars mapping + defs list
        local_count = 0
        for def_msg in ordered_defs:
            for name in uid_to_local_name[def_msg.uid]:
                self.log(f"transcoding local def parent {name}|{local_count} ... input msg:")
                self.log_msg(def_msg)
                assert isinstance(def_msg, DefUniMsg), (
                    f"Expected DefUniMsg in warpc_locals, got {def_msg}"
                )
                if def_msg.uid not in self.defn_ctx:
                    self.defn_ctx[def_msg.uid] = def_msg
                synths: list[DefUniMsg] = []
                native_def = uni_to_py_def(def_msg, self.defn_ctx, name=name, synthetic_defs=synths)
                self.log("primary output msg:")
                self.log_msg(native_def)
                # Add synths first
                for syn in synths:
                    self.log(f"enqueuing synthetic def:\n{syn}")
                    def_msgs.append(uni_to_py_def(syn, self.defn_ctx))
                # Then parent
                def_msgs.append(native_def)
                # Set variable to a reference to the defined resource (with order marker)
                marked_name = name + ORDER_MARKER + str(local_count)
                local_count += 1
                transcoded_vars[marked_name] = native_def.as_ref
                self.log("defined variable:", marked_name, "->", native_def.as_ref)

        # Collect names of non-top-level definitions to hide
        hidden_names: set[str] = set()
        for term_msg in ordered_defs:
            if hasattr(term_msg, 'payload') and hasattr(term_msg.payload, 'is_top_level'):
                if not term_msg.payload.is_top_level:
                    for name in uid_to_local_name[term_msg.uid]:
                        hidden_names.add(name)

        # Add locals that are defless terms
        for name, msg in locals_without_uid.items():
            self.log("transcoding local term:", name, "|", local_count, " input:")
            self.log_msg(msg)

            native_msg = uni_to_py_term(msg, self.defn_ctx)

            self.log("transcoding local term:", name, "|", local_count, " output:")
            self.log_msg(native_msg)

            marked_name = name + ORDER_MARKER + str(local_count)
            local_count += 1
            transcoded_vars[marked_name] = native_msg
            self.log("marked_name:", marked_name)

        self.log("processed local count: ", len(transcoded_vars))
        self.log("hidden names: ", hidden_names)

        # Build VarsMsg with collected vars and defs; attach hidden names for REPL to hide
        if hidden_names:
            try:
                hidden_list = tuple(StrMsg(v=name) for name in hidden_names)
                transcoded_vars['__hidden_names'] = ListMsg(vs=hidden_list)
                self.log("hidden names:", hidden_names)
            except Exception:
                pass
        vars_msg = VarsMsg(vars=transcoded_vars, defs=tuple(def_msgs))
        return vars_msg.to_msgpack()

    def transcode_to_py(self, uni_msg: RpcUniMsg) -> RPCMsg:
        self.log("transcode_to_py incoming")
        self.log_msg(uni_msg)

        self._msg_counter += 1

        # Sanitize names from TypeScript by mangling TS identifiers to Python identifiers
        uni_msg = msg_uni_name_sanitization(uni_msg, self.mangler)

        # Update context _before_ transcoding
        if isinstance(uni_msg, (RequestUniMsg, ResUniMsg)):
            for defmsg in uni_msg.defs:
                self.defn_ctx[defmsg.uid] = defmsg
        if isinstance(uni_msg, FutureCompletedUniMsg):
            self.log("updating context from future completed defs")
            if hasattr(uni_msg.result, 'defs') and isinstance(uni_msg.result, ResUniMsg):
                self.log("found defs in future completed result")
                for defmsg in uni_msg.result.defs:
                    self.defn_ctx[defmsg.uid] = defmsg

        # Transcode
        py_msg = uni_to_py_rpc(uni_msg, self.defn_ctx)
        self.log("transcode_to_py outgoing")
        self.log(py_msg)
        return py_msg

    def transcode_to_uni(self, py_msg: RPCMsg) -> RpcUniMsg:
        self.log("transcode_to_uni incoming")
        self.log(py_msg)

        self._msg_counter += 1

        # Context will be updated while transcoding
        uni_msg = py_to_uni_rpc(py_msg, self.defn_ctx)

        # Restore names for TypeScript by demangling Python identifiers back to TS identifiers
        uni_msg = msg_TS_name_restoration(uni_msg, self.mangler)

        self.log("transcode_to_uni outgoing")
        self.log_msg(uni_msg)
        return uni_msg

    def intercept_sdk(
        self, recv_from_sdk: AsyncRecvBytes, send_to_sdk: AsyncSendBytes
    ) -> tuple[AsyncRecvBytes, AsyncSendBytes]:
        self._orig_recv_bytes = recv_from_sdk
        self._orig_send_bytes = send_to_sdk

        async def transcoded_recv_from_sdk() -> bytes:
            assert self._orig_recv_bytes is not None, "Transcoder recv_bytes is not set"
            warpc_msg = await self._orig_recv_bytes()
            rpc_msg = json_to_rpc_uni(warpc_msg)
            py_msg = self.transcode_to_py(rpc_msg)
            return py_msg.to_msgpack()

        async def transcoded_send_to_sdk(data: bytes) -> None:
            assert self._orig_send_bytes is not None, "Transcoder send_bytes is not set"
            rpc_msg = RPCMsg.from_msgpack(data)
            uni_msg = self.transcode_to_uni(rpc_msg)
            uni_msg_json = uni_to_json(uni_msg)
            await self._orig_send_bytes(uni_msg_json)

        return transcoded_recv_from_sdk, transcoded_send_to_sdk



# fmt: off

from .__ import *
from .base import Msg


__all__ = [
    'DefinitionMsg',
]


################################################################################

if TYPE_CHECKING:
    from .resource_data import ResourceDataMsg
    from .term_resource import UserResourceMsg

################################################################################

class DefinitionMsg(Msg, tag='def'):
    """ABC for messages describing or referencing previously described resource terms."""

    rid: GlobalRID
    data: 'ResourceDataMsg'

    @property
    def as_ref(self) -> 'UserResourceMsg':
        from .term_resource import UserResourceMsg
        return UserResourceMsg(self.rid)

    def __shape__(self) -> str:
        return f_grid(self.rid) + ', ' + self.data.shape

    def repr(self, deep: bool = True) -> str:
        f_data = self.data.repr(False)
        return f'DefinitionMsg({f_grid(self.rid)}, {f_data})'

    def __rich_repr__(self):
        yield f_grid(self.rid)
        yield self.data

    ############################################################################

    def __trans_def_kind__(self) -> Kind:
        return self.data.KIND

    def __trans_int_uid__(self) -> int:
        raise self.rid[2]



# fmt: off

from agentica_internal.warpc.messages import *



def verify_pair(old: Msg, new: Msg):
    assert old.upgrade() == new, f"{old.upgrade()=}  !=  {new}"
    assert new.downgrade() == old, f"{new.downgrade()=}  !=  {old}"


def verify_legacy_conversion():
    request_data = ResourceHasAttrMsg(obj=UserResourceMsg(rid=(0, 1, 2)), attr='foo')
    response_data = ValueMsg(NumberMsg(v=True))

    old_request_msg = legacy.LegacyResourceRequestMsg(mid=3, fid=4, pid=4, info=request_data)
    new_request_msg = FramedRequestMsg(mid=3, fid=4, data=request_data)

    verify_pair(old_request_msg, new_request_msg)

    old_response_msg = legacy.LegacyResourceReplyMsg(mid=3, fid=4, pid=4, info=response_data)
    new_response_msg = FramedResponseMsg(mid=3, fid=4, data=response_data)

    verify_pair(old_response_msg, new_response_msg)

    old_mfi_msg = legacy.LegacyMFIReplyMsg.make(response_data)
    new_mfi_msg = FutureResultMsg(fid=-1, data=response_data)
    verify_pair(old_mfi_msg, new_mfi_msg)


if __name__ == '__main__':
    verify_legacy_conversion()



# fmt: off

from .__ import *
from .rpc_request import RequestMsg


__all__ = [
    'ResourceRequestMsg',
    'ResourceNewMsg',
    'ResourceCallRequestMsg',
    'ResourceCallFunctionMsg',
    'ResourceCallMethodMsg',
    'ResourceCallSystemMethodMsg',
    'ResourceAttrRequestMsg',
    'ResourceHasAttrMsg',
    'ResourceGetAttrMsg',
    'ResourceSetAttrMsg',
    'ResourceDelAttrMsg',
]


################################################################################

if TYPE_CHECKING:
    from ..request.request_resource import *
    from .term import TermMsg
    from .term_resource import ResourceMsg, SystemResourceMsg

################################################################################

class ResourceRequestMsg(RequestMsg):
    """ABC for messages that describe requests of virtual resources.

    These correspond 1-to-1 with the class hierarchy in `warpc.request`.
    """

    LOG_TAGS = 'RSRC'

    def decode(self, dec: DecoderP) -> 'ResourceRequest': ...


################################################################################

class ResourceCallRequestMsg(ResourceRequestMsg):
    """ABC for function-call-like requests; these have a pos and key field."""
    ...


################################################################################

class ResourceNewMsg(ResourceCallRequestMsg, tag='new'):
    """Describes a call to create an instance of a virtualized class."""

    cls: 'ResourceMsg'
    pos: 'Tup[TermMsg]' = ()
    key: 'Rec[TermMsg]' = {}

    def __shape__(self) -> str:
        return self.cls.shape

    def decode(self, dec) -> 'ResourceNew':
        from ..request.request_resource import ResourceNew
        return ResourceNew(
            dec.dec_local_resource(self.cls),
            dec.dec_args(self.pos),
            dec.dec_kwargs(self.key),
        )


################################################################################

class ResourceCallFunctionMsg(ResourceCallRequestMsg, tag='call'):
    """Describes a call to a function."""

    fun: 'ResourceMsg'
    pos: 'Tup[TermMsg]' = ()
    key: 'Rec[TermMsg]' = {}

    def __shape__(self) -> str:
        return self.fun.shape

    def decode(self, dec) -> 'ResourceCallFunction':
        from ..request.request_resource import ResourceCallFunction
        return ResourceCallFunction(
            dec.dec_local_resource(self.fun),
            dec.dec_args(self.pos),
            dec.dec_kwargs(self.key),
        )


################################################################################

class ResourceCallMethodMsg(ResourceCallRequestMsg, tag='callmethod'):
    """Describes a call to a named method of an object."""

    obj: 'ResourceMsg'
    mth:  Name
    pos: 'Tup[TermMsg]' = ()
    key: 'Rec[TermMsg]' = {}

    def __shape__(self) -> str:
        return f'{self.obj.shape},{self.mth!r}'

    def decode(self, dec) -> 'ResourceCallMethod':
        from ..request.request_resource import ResourceCallMethod
        return ResourceCallMethod(
            dec.dec_local_resource(self.obj),
            self.mth,
            dec.dec_args(self.pos),
            dec.dec_kwargs(self.key),
        )


################################################################################

class ResourceCallSystemMethodMsg(ResourceCallRequestMsg, tag='callsys'):
    """Describes a call like `hash(obj)` or `str(obj)` that invokes a special
    dunder method on the object."""

    obj: 'ResourceMsg'
    fun: 'SystemResourceMsg'

    def __shape__(self) -> str:
        return f'{self.obj.shape}, {self.fun.shape}'

    def decode(self, dec) -> 'ResourceCallSystemMethod':
        from ..request.request_resource import ResourceCallSystemMethod
        return ResourceCallSystemMethod(
            dec.dec_resource(self.obj),
            dec.dec_system_resource(self.fun),
        )


################################################################################

class ResourceAttrRequestMsg(ResourceRequestMsg):
    """ABC for request messages referring to attributes."""

    obj: 'ResourceMsg'
    attr: Name

    def __shape__(self) -> str:
        return f'{self.obj.shape},{self.attr!r}'


################################################################################

class ResourceHasAttrMsg(ResourceAttrRequestMsg, tag='hasattr'):

    def decode(self, dec) -> 'ResourceHasAttr':
        from ..request.request_resource import ResourceHasAttr
        return ResourceHasAttr(
            dec.dec_local_resource(self.obj),
            self.attr,
        )


################################################################################

class ResourceGetAttrMsg(ResourceAttrRequestMsg, tag='getattr'):

    def decode(self, dec) -> 'ResourceGetAttr':
        from ..request.request_resource import ResourceGetAttr
        return ResourceGetAttr(
            dec.dec_local_resource(self.obj),
            self.attr,
        )


################################################################################

class ResourceDelAttrMsg(ResourceAttrRequestMsg, tag='delattr'):

    def decode(self, dec) -> 'ResourceDelAttr':
        from ..request.request_resource import ResourceDelAttr
        return ResourceDelAttr(
            dec.dec_local_resource(self.obj),
            self.attr,
        )


################################################################################

class ResourceSetAttrMsg(ResourceAttrRequestMsg, tag='setattr'):

    val: 'TermMsg'

    def decode(self, dec) -> 'ResourceSetAttr':
        from ..request.request_resource import ResourceSetAttr
        return ResourceSetAttr(
            dec.dec_local_resource(self.obj),
            self.attr,
            dec.dec_any(self.val),
        )



# fmt: off

from typing import ClassVar, TypeGuard

from .hashing import raw_str_hash

__all__ = [
    'Sentinel',
    'NoValue',
    'FalseSentinel',
    'ArgDefault',
    'FieldAbsent',
    'Pending',
    'Closed',
    'Canceled',
    'Errored',
    'NO_VALUE',
    'ARG_DEFAULT',
    'FIELD_ABSENT',
    'CLOSED',
    'PENDING',
    'CANCELED',
    'ERRORED',
    'SENTINEL_TYPES',
    'SENTINEL_IDS',
    'SENTINEL_OBJS',
    'SENTINELS',
    'is_sentinel',
]

###############################################################################

SENTINELS: dict[str, 'Sentinel'] = {}

SENTINEL_TYPES = set()
SENTINEL_IDS = set()
SENTINEL_OBJS = set()

class Sentinel:
    __slots__ = ('__hash_value__',)

    hval: ClassVar[int]
    name: ClassVar[str]
    inst: ClassVar['Sentinel']

    def __init_subclass__(cls, name: str):
        cls.name = name
        cls.inst = inst = object.__new__(cls)
        cls.hval = raw_str_hash(name) ^ 0x1234
        inst.__hash_value__ = cls.hval
        SENTINEL_TYPES.add(cls)
        SENTINEL_IDS.add(id(inst))
        SENTINEL_OBJS.add(inst)
        SENTINELS[name] = inst

    def __new__(cls):
        return cls.inst

    def __repr__(self) -> str:
        return type(self).name

    def __hash__(self) -> int:
        return self.__hash_value__

    def __eq__(self, other) -> bool:
        return self is other

    def __ne__(self, other) -> bool:
        return self is not other

    __short_str__ = __repr__


###############################################################################

class FalseSentinel(Sentinel, name=''):

    def __bool__(self) -> bool:
        return False

###############################################################################

class NoValue(FalseSentinel,     name='NO_VALUE'): ...
class ArgDefault(FalseSentinel,  name='...'): ...
class FieldAbsent(FalseSentinel, name='FIELD_ABSENT'): ...
class Pending(FalseSentinel,     name='PENDING'): ...
class Closed(FalseSentinel,      name='CLOSED'): ...
class Canceled(FalseSentinel,    name='CANCELED'): ...
class Errored(FalseSentinel,     name='ERRORED'): ...

NO_VALUE     = NoValue()
ARG_DEFAULT  = ArgDefault()
FIELD_ABSENT = FieldAbsent()
PENDING      = Pending()
CLOSED       = Closed()
CANCELED     = Canceled()
ERRORED      = Errored()


###############################################################################

# this doesn't work, for some reason...
# def none_schema(cls, source_type, handler):
#     from pydantic_core import core_schema
#     return core_schema.none_schema()
#
# ArgDefault.__get_pydantic_core_schema__ = classmethod(none_schema)
#
# ensure that virtualized functions with foo=ARG_DEFAULT are understood by
# pydantic

###############################################################################

is_sentinel_id = SENTINEL_IDS.__contains__


def is_sentinel(obj) -> TypeGuard[Sentinel]:
    return is_sentinel_id(id(obj))



from typing import Protocol as TypingProtocol

from msgspec import UnsetType

from agentica_internal.warpc.alias import *
from agentica_internal.warpc.messages import *
from agentica_internal.warpc.msg.resource_data import (
    ClassDataMsg,
    ForwardRefDataMsg,
    FunctionDataMsg,
    FutureDataMsg,
    GenericAliasDataMsg,
    TypeUnionDataMsg,
)
from agentica_internal.warpc.msg.rpc_event import FutureCanceledMsg as PyFutureCanceledMsg
from agentica_internal.warpc.msg.rpc_event import FutureCompletedMsg as PyFutureCompletedMsg
from agentica_internal.warpc.msg.rpc_legacy import LegacyResourceReplyMsg, LegacyResourceRequestMsg
from agentica_internal.warpc.msg.term_atom import ForwardRefTypeMsg
from agentica_internal.warpc.msg.term_exception import ExceptionMsg
from agentica_internal.warpc.msg.term_resource import SystemResourceMsg, UserResourceMsg
from agentica_internal.warpc.system import LRID_TO_SRID
from agentica_internal.warpc_transcode.conv_utils import get_def_or_sysref_from_ctx
from agentica_internal.warpc_transcode.uni_msgs import DefUniMsg
from agentica_internal.warpc_transcode.uni_sys_id import ATOMIC_IDS

from .conv_utils import *
from .uni_msgs import *
from .uni_msgs import (
    EventUniMsg,
    ForeignExceptionUniMsg,
    FutureCanceledUniMsg,
    FutureCompletedUniMsg,
    GenericExceptionUniMsg,
    InternalErrorUniMsg,
    toMethodSignature,
)

__all__ = ['uni_to_py']

"""
note: also see warpc/msg/__final__.py for how message classes are finalized...
"""


EXCLUDE_FROM_BASES = {BUILTIN_UNI_IDS["Object"]}


# Uni Any -> Py Any
def uni_to_py(msg: UniMsg, ctx: dict[DefnUID, DefUniMsg]) -> Msg:
    if isinstance(msg, RpcUniMsg):
        return uni_to_py_rpc(msg, ctx)
    elif isinstance(msg, ConceptUniMsg):
        return uni_to_py_term(msg, ctx)
    else:
        raise ValueError(f'Unsupported universal message: {msg}')


# Uni RPC -> Py RPC
def uni_to_py_rpc(msg: RpcUniMsg, ctx: dict[DefnUID, DefUniMsg]) -> RPCMsg:
    if isinstance(msg, RequestUniMsg):
        # HACK: use requestedFID for mid
        pid, fid, mid = uni_to_py_frame_request(msg)

        info = uni_to_py_request(msg, ctx)
        defs = tuple(uni_to_ordered_py_defs(msg.defs, ctx))

        return LegacyResourceRequestMsg(
            mid=mid,
            fid=fid,
            pid=pid,
            info=info,
            defs=defs,
        ).upgrade()  # type: ignore[arg-type]

    elif isinstance(msg, ResponseUniMsg):
        # HACK: mid is not known in uni response (set to fid)
        pid, fid = uni_to_py_frame_response(msg)
        info = uni_to_py_response(msg, ctx)
        if hasattr(msg, 'defs'):
            assert isinstance(msg, ResUniMsg), f"Got unexpected defs in response"
            defs = tuple(uni_to_ordered_py_defs(msg.defs, ctx))
        else:
            defs = ()
        return LegacyResourceReplyMsg(mid=fid, fid=fid, pid=pid, info=info, defs=defs).upgrade()  # type: ignore[arg-type]

    elif isinstance(msg, EventUniMsg):
        if isinstance(msg, FutureCanceledUniMsg):
            future_id = msg.future_id
            return PyFutureCanceledMsg(future_id=future_id)

        if isinstance(msg, FutureCompletedUniMsg):
            future_id = msg.future_id
            if hasattr(msg.result, 'defs'):
                assert isinstance(msg.result, ResUniMsg), f"Got unexpected defs in completed result"
                defs = tuple(uni_to_ordered_py_defs(msg.result.defs, ctx))
            else:
                defs = ()
            result_msg = uni_to_py_response(msg.result, ctx)
            return PyFutureCompletedMsg(future_id=future_id, result=result_msg, defs=defs)

        raise ValueError(f'Unknown EventUniMsg: {msg}')

    else:
        raise ValueError(f'Unknown RpcUniMsg: {msg}')


# Uni Request -> Py Request
def uni_to_py_request(msg: RequestUniMsg, ctx: dict[DefnUID, DefUniMsg]) -> ResourceRequestMsg:
    if isinstance(msg, CallNewUniMsg):
        cls = uni_to_py_term(msg.payload.cls, ctx)
        assert isinstance(cls, ResourceMsg), f"Expected ResourceMsg for cls, got {cls}"
        pos, key = make_py_call_args(msg.payload.args, ctx)
        info = ResourceNewMsg(cls=cls, pos=pos, key=key)
        return info

    if isinstance(msg, CallFunctionUniMsg):
        fun = uni_to_py_term(msg.payload.fun, ctx)
        assert isinstance(fun, ResourceMsg), f"Expected ResourceMsg for fun, got {fun}"
        pos, key = make_py_call_args(msg.payload.args, ctx)
        info = ResourceCallFunctionMsg(fun=fun, pos=pos, key=key)
        return info

    if isinstance(msg, CallMethodUniMsg):
        owner = uni_to_py_term(msg.payload.owner, ctx)
        assert isinstance(owner, ResourceMsg), f"Expected ResourceMsg for owner, got {owner}"
        pos, key = make_py_call_args(msg.payload.args, ctx)
        info = ResourceCallMethodMsg(obj=owner, mth=msg.payload.method_name, pos=pos, key=key)
        return info

    if isinstance(msg, HasAttrUniMsg):
        ref = uni_to_py_term(msg.payload.owner, ctx)
        assert isinstance(ref, ResourceMsg), f"Expected ResourceMsg for hasattr ref, got {ref}"
        info = ResourceHasAttrMsg(obj=ref, attr=msg.payload.attr)
        return info

    if isinstance(msg, GetAttrUniMsg):
        ref = uni_to_py_term(msg.payload.owner, ctx)
        assert isinstance(ref, ResourceMsg), f"Expected ResourceMsg for getattr ref, got {ref}"
        info = ResourceGetAttrMsg(obj=ref, attr=msg.payload.attr)
        return info

    if isinstance(msg, SetAttrUniMsg):
        ref = uni_to_py_term(msg.payload.owner, ctx)
        assert isinstance(ref, ResourceMsg), f"Expected ResourceMsg for setattr ref, got {ref}"
        val = uni_to_py_term(msg.payload.val, ctx)
        info = ResourceSetAttrMsg(obj=ref, attr=msg.payload.attr, val=val)
        return info

    if isinstance(msg, DelAttrUniMsg):
        ref = uni_to_py_term(msg.payload.owner, ctx)
        assert isinstance(ref, ResourceMsg), f"Expected ResourceMsg for delattr ref, got {ref}"
        info = ResourceDelAttrMsg(obj=ref, attr=msg.payload.attr)
        return info

    raise ValueError(f'Unsupported universal request: {msg}')


# Uni Response -> Py Response
def uni_to_py_response(msg: ResponseUniMsg, ctx: dict[DefnUID, DefUniMsg]) -> ResultMsg:
    if isinstance(msg, OkUniMsg):
        return OK_MSG

    if isinstance(msg, ResUniMsg):
        term = uni_to_py_term(msg.payload.result, ctx)
        return ValueMsg(val=term)

    if isinstance(msg, ErrUniMsg):
        excp_msg = uni_to_py_term(msg.payload.error, ctx)
        return ErrorMsg(exc=excp_msg)

    raise ValueError(f'Unsupported universal response: {msg}')


# Uni Concept -> Py Term
def uni_to_py_term(term: ConceptUniMsg, ctx: dict[DefnUID, DefUniMsg]) -> TermMsg:
    if isinstance(term, RefUniMsg) and (
        not isinstance(term, TypeStringRefUniMsg)
        or term.uid.resource < 0  # isinstance TypeStringRefUniMsg and uid.resource < 0
    ):
        rid = term.uid.resource

        # If the universal resource id is a negative known system id, map to system rid
        if rid < 0:
            from .conv_utils import uniSysIdToSrid

            sys_id = uniSysIdToSrid(rid)
            if sys_id is not None:
                return SystemResourceMsg(sid=sys_id)

        if term.system:
            return SystemResourceMsg(sid=rid)

        rid = resourceUIDToGrid(term.uid)
        return UserResourceMsg(rid=rid)

    if isinstance(term, DefUniMsg):
        # For terms, reference the resource by RID; definitions go separately in defs
        return UserResourceMsg(rid=resourceUIDToGrid(term.uid))

    # Atomics
    if isinstance(term, NoneUniMsg):
        return NoneMsg.MSG
    if isinstance(term, BoolUniMsg):
        return NumberMsg(v=term.val)
    if isinstance(term, IntUniMsg):
        return NumberMsg(v=term.val)
    if isinstance(term, FloatUniMsg):
        return NumberMsg(v=term.val)
    if isinstance(term, StrUniMsg):
        return StrMsg(v=term.val)
    if isinstance(term, TypeStringRefUniMsg):
        return ForwardRefTypeMsg(v=term.name)
    if isinstance(term, BytesUniMsg):
        return BytesMsg(v=term.val)

    # Symbols
    if isinstance(term, EnumValUniMsg):
        return EnumValMsg(val=uni_to_py_term(term.val, ctx), cls=uni_to_py_term(term.cls, ctx))  # type: ignore[call-arg]

    # Containers
    if isinstance(term, ArrayUniMsg):
        vs = tuple(uni_to_py_term(v, ctx) for v in term.val)
        return ListMsg(vs=vs)
    if isinstance(term, SetUniMsg):
        vs = tuple(uni_to_py_term(v, ctx) for v in term.val)
        return SetMsg(vs=vs)
    if isinstance(term, MapUniMsg):
        ks = tuple(uni_to_py_term(k, ctx) for k in term.val.keys)
        vs = tuple(uni_to_py_term(v, ctx) for v in term.val.vals)
        return DictMsg(ks=ks, vs=vs)
    if isinstance(term, TupleUniMsg):
        vs = tuple(uni_to_py_term(v, ctx) for v in term.val)
        return TupleMsg(vs=vs)

    # Exceptions
    if isinstance(term, ForeignExceptionUniMsg):
        return ExceptionMsg(
            cls=uni_to_py_term(term.excp_cls, ctx),
            args=tuple(uni_to_py_term(v, ctx) for v in term.excp_args),
            name='ForeignException',
            loc=None,
            stack=[],
        )  # type: ignore[arg-type]
    if isinstance(term, InternalErrorUniMsg):
        return ExceptionMsg(
            cls=SystemResourceMsg(sid=LRID_TO_SRID.get(id(Exception), 0)),
            args=(StrMsg(v=term.error),),
            name='Exception',
            loc=None,
            stack=[],
        )  # type: ignore[arg-type]
    if isinstance(term, GenericExceptionUniMsg):
        return ExceptionMsg(
            cls=SystemResourceMsg(sid=LRID_TO_SRID.get(id(Exception), 0)),
            args=tuple(StrMsg(v=v) for v in term.excp_str_args),
            name=term.excp_cls_name,
            loc=None,
            stack=list(v for v in term.excp_stack),
        )  # type: ignore[arg-type]

    raise ValueError(f'Unknown universal term: {term}')


# Uni Def Msg -> Py Definition Msg
def uni_to_py_def(
    defmsg: DefUniMsg,
    ctx: dict[DefnUID, DefUniMsg],
    name: str | None = None,
    synthetic_defs: list[DefUniMsg] | None = None,
) -> DefinitionMsg:
    rid = resourceUIDToGrid(defmsg.uid)

    if isinstance(defmsg, ClassDefUniMsg):
        if defmsg.payload.system_resource:
            sys_data = TypeUnionDataMsg(alts=(SystemResourceMsg(sid=defmsg.uid.resource),))  # type: ignore[arg-type]
            return DefinitionMsg(rid=rid, data=sys_data)

        # Case of "function types" like ((x: int) -> string)
        if isinstance(callmsg := defmsg.payload.supplied_type_args, FunctionDefUniMsg):
            # Comment:
            # * The current approach construct a callable class to encode the function type
            # * but we could also easily turn this into a `Callable[[A,B],C]` GenericAlias instead of a class

            # Synthesize __call__
            start_uid = make_space_for_uid(defmsg.uid.resource)
            call_uid = DefnUID(world='client', resource=start_uid - 1)

            # Create a new function def with the unique UID
            # Copy the payload fields from the original function
            call_payload = FunctionPayload(
                name="__call__",
                arguments=callmsg.payload.arguments,
                doc=callmsg.payload.doc if callmsg.payload.doc is not UNSET else UNSET,
                returnType=callmsg.payload.returnType
                if callmsg.payload.returnType is not UNSET
                else UNSET,
                module=callmsg.payload.module if callmsg.payload.module is not UNSET else UNSET,
            )
            call_func = FunctionDefUniMsg(uid=call_uid, payload=call_payload)

            # Register in context and create reference
            ctx[call_uid] = call_func
            call_ref = UserResourceMsg(rid=resourceUIDToGrid(call_uid))

            if synthetic_defs is not None:
                synthetic_defs.append(call_func)

            data = class_payload_to_py_data(
                defmsg.payload, defmsg.uid, ctx, call_ref, synthetic_defs
            )
            return DefinitionMsg(rid=rid, data=data)

        if (sys_msg := uni_def_to_py_sys(defmsg, ctx)) is not None:
            return sys_msg

        data = class_payload_to_py_data(defmsg.payload, defmsg.uid, ctx, None, synthetic_defs)
        return DefinitionMsg(rid=rid, data=data)

    if isinstance(defmsg, FunctionDefUniMsg):
        data = function_payload_to_py_data(defmsg.payload, ctx)
        return DefinitionMsg(rid=rid, data=data)

    if isinstance(defmsg, ObjectDefUniMsg):
        if defmsg.payload.cls is UNSET:
            raise ValueError("Cannot transcode object without cls")

        ## the special future shortcut
        if (
            (cls_def := get_def_or_sysref_from_ctx(ctx, defmsg.payload.cls)) is not None
            and isinstance(cls_def, ClassDefUniMsg)
            and cls_def.payload.instance_of_generic is not UNSET
            and cls_def.payload.instance_of_generic.resource == BUILTIN_UNI_IDS["Future"]
        ):
            return DefinitionMsg(
                rid=rid, data=FutureDataMsg(future=defmsg.uid.resource, result=None)
            )  # type: ignore[call-arg]

        ## non-future objects
        data = object_payload_to_py_data(
            defmsg.payload if defmsg.payload is not UNSET else ObjectPayload(),
            ctx,
        )
        return DefinitionMsg(rid=rid, data=data)

    if isinstance(defmsg, MethodSignatureUniMsg):
        # Represent as a function definition in native with class context
        memberof = get_def_or_sysref_from_ctx(ctx, defmsg.methodOf)
        assert isinstance(memberof, (ClassDefUniMsg, InterfaceUniMsg, IntersectionUniMsg))
        data = function_payload_to_py_data(defmsg.payload, ctx, class_def=memberof)
        return DefinitionMsg(rid=rid, data=data)

    # Annotation defs
    if isinstance(defmsg, UnionUniMsg):
        # Map union classes to term alts
        alts = tuple(uni_to_py_term(ref, ctx) for ref in defmsg.payload.classes)
        data = TypeUnionDataMsg(alts=alts)  # type: ignore[arg-type]
        return DefinitionMsg(rid=rid, data=data)

    if isinstance(defmsg, IntersectionUniMsg):
        data = intersection_to_class_data(
            defmsg, rid, ctx, name=name, synthetic_defs=synthetic_defs
        )
        return DefinitionMsg(rid=rid, data=data)

    if isinstance(defmsg, InterfaceUniMsg):
        data = class_payload_to_protocol_data(defmsg.payload, defmsg.uid, ctx, synthetic_defs)
        return DefinitionMsg(rid=rid, data=data)

    if isinstance(defmsg, PlaceholderUniMsg):
        raise NotImplementedError('Placeholder defs are not supported')

    raise ValueError(f'Unsupported universal def: {defmsg}')


# uni def -> py sys (or union thereof)
def uni_def_to_py_sys(defmsg: DefUniMsg, ctx: dict[DefnUID, DefUniMsg]) -> DefinitionMsg | None:
    def make_sys_msg(py_ids: list[int]) -> DefinitionMsg:
        alts = tuple(SystemResourceMsg(sid=py_id) for py_id in py_ids)
        union = TypeUnionDataMsg(alts=alts)  # type: ignore[arg-type]
        return DefinitionMsg(rid=resourceUIDToGrid(defmsg.uid), data=union)

    if defmsg.uid.resource == ATOMIC_IDS["object"]:
        py_ids = [LRID_TO_SRID.get(id(object), 0)]
        return make_sys_msg(py_ids)
    if defmsg.uid.resource == ATOMIC_IDS["string"]:
        py_ids = [LRID_TO_SRID.get(id(str), 0)]
        return make_sys_msg(py_ids)
    if defmsg.uid.resource == ATOMIC_IDS["number"]:
        py_ids = [LRID_TO_SRID.get(id(int), 0), LRID_TO_SRID.get(id(float), 0)]
        return make_sys_msg(py_ids)
    if defmsg.uid.resource == ATOMIC_IDS["boolean"]:
        py_ids = [LRID_TO_SRID.get(id(bool), 0)]
        return make_sys_msg(py_ids)
    if defmsg.uid.resource == ATOMIC_IDS["undefined"]:
        py_ids = [LRID_TO_SRID.get(id(type(None)), 0)]
        return make_sys_msg(py_ids)
    return None


# Uni Intersection -> Py Class Data
def intersection_to_class_data(
    defmsg: IntersectionUniMsg,
    rid: GlobalRID,
    ctx: dict[DefnUID, DefUniMsg],
    name: str | None = None,
    synthetic_defs: list[DefUniMsg] | None = None,
) -> ClassDataMsg:
    type_sid = LRID_TO_SRID.get(id(type), 0)
    cls_msg = SystemResourceMsg(sid=type_sid)

    base_names: list[str] = []
    base_uni_refs: list[RefUniMsg] = []
    module: str | None = None
    for ref in defmsg.payload.classes:
        base_def = get_def_or_sysref_from_ctx(ctx, ref)
        if base_def and isinstance(base_def, (ClassDefUniMsg, InterfaceUniMsg)):
            # Exclude builtin types from bases
            if (
                base_def.payload.instance_of_generic is not UNSET
                and base_def.payload.instance_of_generic.resource in EXCLUDE_FROM_BASES
            ):
                continue

            base_names.append(base_def.payload.name)  # type: ignore[attr-defined]
            base_uni_refs.append(ref)  # Add the ref to the list of bases
            # Use the module from the first base that has one

            if (
                module is None
                and hasattr(base_def.payload, 'module')
                and type(base_def.payload.module) is str
            ):  # type: ignore[attr-defined]
                module = base_def.payload.module  # type: ignore[attr-defined]

    if name is None:
        name = '_Also_'.join(base_names) if base_names else f'Intersection_{rid[2]}'

    bases = tuple(uni_to_py_term(ref, ctx) for ref in base_uni_refs)

    # Synthesize uninstantiable __init__ with Never parameter
    init_ref = add_uninstantiable_init_method(name, defmsg.uid, ctx, synthetic_defs)
    methods: dict[str, tuple[str, ResourceMsg]] = (
        {'__init__': ('instance', init_ref)} if init_ref else {}
    )

    return ClassDataMsg(
        name=name,
        cls=cls_msg,
        bases=bases,
        methods=methods,
        qname=name,
        module="__main__",  # TODO: maybe one day module if module is not None else '__main__',
        doc=f'Protocol combining: {", ".join(base_names)}' if base_names else '',
        annos={},
        attrs={},
        sattrs=(),
        keys=(),
    )  # type: ignore[call-arg]


# Uni Class Payload -> Py Class Data
def class_payload_to_py_data(
    data: ClassPayload,
    class_uid: DefnUID,
    ctx: dict[DefnUID, DefUniMsg],
    call_rsrc: ResourceMsg | None = None,
    synthetic_defs: list[DefUniMsg] | None = None,
) -> ResourceDataMsg:
    # Short cut for enums
    if (
        data.instance_of_generic is not UNSET
        and data.instance_of_generic.resource == BUILTIN_UNI_IDS["Enum"]
    ):
        members = {}
        assert isinstance(data.supplied_type_args, list)

        # Determine enum_kind by inspecting all arg types; prefer int->str->any if mixed
        arg_types = {type(arg.type) for arg in data.supplied_type_args}
        if all(t is IntUniMsg for t in arg_types):
            enum_kind = 'int'
        elif all(t is StrUniMsg for t in arg_types):
            enum_kind = 'str'
        else:
            enum_kind = 'any'

        for argMsg in data.supplied_type_args:
            members[argMsg.name] = uni_to_py_term(argMsg.type, ctx)
        return EnumClassDataMsg(name=data.name, members=members, kind=enum_kind)  # type: ignore[call-arg]

    # Short cuts for other generic aliases
    if data.instance_of_generic is not UNSET and call_rsrc is None:
        from .conv_utils import uniSysIdToSrid

        classvar_uid = data.instance_of_generic
        sys_id = uniSysIdToSrid(classvar_uid.resource)
        origin: TermMsg
        if sys_id is not None:
            origin = SystemResourceMsg(sid=sys_id)
        else:
            origin = UserResourceMsg(rid=resourceUIDToGrid(classvar_uid))

        args: list[TermMsg] = []
        if isinstance(data.supplied_type_args, list):
            for arg in data.supplied_type_args:
                args.append(uni_to_py_term(arg.type, ctx))

        return GenericAliasDataMsg(origin=origin, args=tuple(args))  # type: ignore[call-arg]

    # General classes
    type_sid = LRID_TO_SRID.get(id(type), 0)
    cls_msg = SystemResourceMsg(sid=type_sid)

    bases = (
        tuple(uni_to_py_term(b, ctx) for b in (data.bases or [])) if data.bases is not UNSET else ()
    )

    # Synthesize __init__ from constructor arguments
    new_synthetic_init = add_init_method(data, class_uid, ctx)
    if synthetic_defs is not None:
        synthetic_defs.extend(new_synthetic_init)

    # Synthesize dict protocol methods if needed
    new_synthetic = add_dict_protocol_methods(data, class_uid, ctx)
    if synthetic_defs is not None:
        synthetic_defs.extend(new_synthetic)

    # methods: mapping name -> resource (function def/ref)
    methods: dict[str, tuple[str, ResourceMsg]] = {}
    if data.methods is not UNSET:
        for m in data.methods:
            name = m.name
            # skip private methods
            if m.is_private is not UNSET and m.is_private:
                continue
            fun_ref = uni_to_py_term(m.function, ctx)
            assert isinstance(fun_ref, ResourceMsg)
            methods[name] = ('instance', fun_ref)

    annos, _instance_keys, class_keys = make_annos_and_keys(data, class_uid, ctx, synthetic_defs)

    # Only instance fields are instance keys for now
    keys = tuple(class_keys)

    # For callable classes add __call__ (by reference)
    if call_rsrc is not None:
        methods['__call__'] = ('instance', call_rsrc)
        keys = keys + ('__call__',)

    return ClassDataMsg(  # type: ignore[call-arg]
        name=data.name,
        cls=cls_msg,
        bases=tuple(bases),
        methods=methods,
        qname=data.name,
        module="__main__",  # TODO: maybe one day "__main__" if data.module is UNSET else data.module,
        doc="" if data.doc is UNSET else data.doc,
        annos=annos,
        keys=keys,
    )


def make_annos_and_keys(
    data: ClassPayload,
    class_uid: DefnUID,
    ctx: dict[DefnUID, DefUniMsg],
    synthetic_defs: list[DefUniMsg] | None = None,
) -> tuple[dict[str, TermMsg], list[str], list[str]]:
    # annotations to fields
    annos: dict[str, TermMsg] = {}
    instance_keys: list[str] = []
    class_keys: list[str] = []
    if data.fields is not UNSET:
        idx = 0
        for f in data.fields:
            # skip private fields
            if f.is_private is not UNSET and f.is_private:
                continue

            # Instance vs static fields
            is_static = f.is_static is not UNSET and f.is_static
            base_term = uni_to_py_term(f.type, ctx)

            # Instance fields
            if not is_static:
                annos[f.name] = base_term
                instance_keys.append(f.name)
                continue

            # Class fields
            class_keys.append(f.name)
            classvar_wrapper = try_make_classvar(f.type, class_uid, ctx, synthetic_defs, offset=idx)
            if classvar_wrapper is not None:
                annos[f.name] = classvar_wrapper
                idx += 1
            else:
                annos[f.name] = base_term
    return annos, instance_keys, class_keys


def make_space_for_uid(uid_in: int) -> int:
    return -10000 - uid_in * 100


def try_make_classvar(
    field_type: RefUniMsg,
    class_uid: DefnUID,
    ctx: dict[DefnUID, DefUniMsg],
    synthetic_defs: list[DefUniMsg] | None = None,
    offset: int = 0,
) -> UserResourceMsg:
    # Wrap static field types in typing.ClassVar[T]
    classvar_uni_id = BUILTIN_UNI_IDS["ClassVar"]
    classvar_defn_id = DefnUID(world='client', resource=classvar_uni_id)

    # Allocate a unique UID for this alias
    start_uid = make_space_for_uid(class_uid.resource)
    alias_uid = DefnUID(world='client', resource=start_uid - 9 - offset)
    alias_payload = ClassPayload(
        name='GenericAlias',
        fields=[],
        methods=[],
        instance_of_generic=classvar_defn_id,
        supplied_type_args=[TypeArgument(name='a0', type=field_type)],
    )
    alias_uni_def = ClassDefUniMsg(uid=alias_uid, payload=alias_payload)

    # Register synthetic def for output and referencing
    ctx[alias_uid] = alias_uni_def
    if synthetic_defs is not None:
        synthetic_defs.append(alias_uni_def)

    return UserResourceMsg(rid=resourceUIDToGrid(alias_uid))


def add_init_method(
    data: ClassPayload, class_uid: DefnUID, ctx: dict[DefnUID, DefUniMsg]
) -> list[MethodSignatureUniMsg]:
    if data.ctor_args is UNSET or not data.ctor_args:
        return []

    start_uid = make_space_for_uid(class_uid.resource)
    init_uid = DefnUID(world='client', resource=start_uid - 0)

    if init_uid in ctx:
        return []

    none_uid = DefnUID(world='client', resource=BUILTIN_UNI_IDS["None"])
    none_ref = RefUniMsg(uid=none_uid, system=True)

    init_payload = FunctionPayload(
        name='__init__',
        arguments=data.ctor_args,
        returnType=none_ref,
        doc=f"Initialize an instance of {data.name}.",
    )

    init_func = FunctionDefUniMsg(uid=init_uid, payload=init_payload)
    membership = Membership(uid=class_uid, kind='instance')
    init_method = toMethodSignature(init_func, membership)

    ctx[init_uid] = init_method

    if data.methods is UNSET:
        data.methods = []
    data.methods.append(ClassMethod(name='__init__', function=RefUniMsg(uid=init_uid)))

    return [init_method]


def add_uninstantiable_init_method(
    class_name: str,
    class_uid: DefnUID,
    ctx: dict[DefnUID, DefUniMsg],
    synthetic_defs: list[DefUniMsg] | None = None,
) -> ResourceMsg | None:
    start_uid = make_space_for_uid(class_uid.resource)
    init_uid = DefnUID(world='client', resource=start_uid - 0)

    if init_uid in ctx:
        return None

    none_uid = DefnUID(world='client', resource=BUILTIN_UNI_IDS["None"])
    none_ref = RefUniMsg(uid=none_uid, system=True)

    never_uid = DefnUID(world='client', resource=BUILTIN_UNI_IDS["Never"])
    never_ref = RefUniMsg(uid=never_uid, system=True)

    init_payload = FunctionPayload(
        name='__init__',
        arguments=[FunctionArgument(name='_', type=never_ref)],
        returnType=none_ref,
        doc=f"Cannot instantiate {class_name}.",
    )

    init_func = FunctionDefUniMsg(uid=init_uid, payload=init_payload)
    membership = Membership(uid=class_uid, kind='instance')
    init_method = toMethodSignature(init_func, membership)

    ctx[init_uid] = init_method
    if synthetic_defs is not None:
        synthetic_defs.append(init_method)

    # Return the __init__ method reference
    init_ref = uni_to_py_term(RefUniMsg(uid=init_uid), ctx)
    assert isinstance(init_ref, ResourceMsg)
    return init_ref


def add_dict_protocol_methods(
    data: ClassPayload, class_uid: DefnUID, ctx: dict[DefnUID, DefUniMsg]
) -> list[MethodSignatureUniMsg]:
    if data.index_signature is UNSET:
        return []

    key_type = data.index_signature.key_type
    value_type = data.index_signature.value_type
    map_type_uid = data.index_signature.map_type.uid.resource

    start_uid = make_space_for_uid(map_type_uid)
    getitem_uid = DefnUID(world='client', resource=start_uid - 1)
    setitem_uid = DefnUID(world='client', resource=start_uid - 2)
    delitem_uid = DefnUID(world='client', resource=start_uid - 3)
    contains_uid = DefnUID(world='client', resource=start_uid - 4)

    # for __contains__
    bool_uid = DefnUID(world='client', resource=BUILTIN_UNI_IDS["Boolean"])
    bool_ref = RefUniMsg(uid=bool_uid, system=True)
    # for __setitem__ and __delitem__
    none_uid = DefnUID(world='client', resource=BUILTIN_UNI_IDS["None"])
    none_ref = RefUniMsg(uid=none_uid, system=True)

    newly_created: list[MethodSignatureUniMsg] = list()
    membership = Membership(uid=class_uid, kind='instance')

    if getitem_uid not in ctx:
        getitem_payload = FunctionPayload(
            name='__getitem__',
            arguments=[
                FunctionArgument(name='key', type=key_type),
            ],
            returnType=value_type,
            doc="Objects of this class are dict-like, and support indexing with `obj[key]`.",
        )
        getitem_func = FunctionDefUniMsg(uid=getitem_uid, payload=getitem_payload)
        getitem_method = toMethodSignature(getitem_func, membership)
        ctx[getitem_uid] = getitem_method
        newly_created.append(getitem_method)

    if setitem_uid not in ctx:
        setitem_payload = FunctionPayload(
            name='__setitem__',
            arguments=[
                FunctionArgument(name='key', type=key_type),
                FunctionArgument(name='value', type=value_type),
            ],
            returnType=none_ref,
            doc="Objects of this class are dict-like, and support assignment with `obj[key] = value`.",
        )
        setitem_func = FunctionDefUniMsg(uid=setitem_uid, payload=setitem_payload)
        setitem_method = toMethodSignature(setitem_func, membership)
        ctx[setitem_uid] = setitem_method
        newly_created.append(setitem_method)

    if delitem_uid not in ctx:
        delitem_payload = FunctionPayload(
            name='__delitem__',
            arguments=[
                FunctionArgument(name='key', type=key_type),
            ],
            returnType=none_ref,
            doc="Objects of this class are dict-like, and support deletion with `del obj[key]`.",
        )
        delitem_func = FunctionDefUniMsg(uid=delitem_uid, payload=delitem_payload)
        delitem_method = toMethodSignature(delitem_func, membership)
        ctx[delitem_uid] = delitem_method
        newly_created.append(delitem_method)

    if contains_uid not in ctx:
        contains_payload = FunctionPayload(
            name='__contains__',
            arguments=[
                FunctionArgument(name='key', type=key_type),
            ],
            returnType=bool_ref,
            doc="Objects of this class are dict-like, and support checking for membership with `key in obj`.",
        )
        contains_func = FunctionDefUniMsg(uid=contains_uid, payload=contains_payload)
        contains_method = toMethodSignature(contains_func, membership)
        ctx[contains_uid] = contains_method
        newly_created.append(contains_method)

    if data.methods is UNSET:
        data.methods = list()

    data.methods.extend(
        [
            ClassMethod(name='__getitem__', function=RefUniMsg(uid=getitem_uid)),
            ClassMethod(name='__setitem__', function=RefUniMsg(uid=setitem_uid)),
            ClassMethod(name='__delitem__', function=RefUniMsg(uid=delitem_uid)),
            ClassMethod(name='__contains__', function=RefUniMsg(uid=contains_uid)),
        ]
    )

    return newly_created


def uni_inherits_from_interface(base_ref: RefUniMsg, ctx: dict[DefnUID, DefUniMsg]) -> bool:
    visited: set[DefnUID] = set()

    def check_uni_base(ref: RefUniMsg) -> bool:
        if ref.uid in visited:
            return False
        visited.add(ref.uid)

        base_def = get_def_or_sysref_from_ctx(ctx, ref)
        if base_def is None:
            return False

        if isinstance(base_def, InterfaceUniMsg):
            return True

        if isinstance(base_def, ClassDefUniMsg):
            if base_def.payload.bases is not UNSET:
                for parent_ref in base_def.payload.bases:
                    if check_uni_base(parent_ref):
                        return True

        return False

    return check_uni_base(base_ref)


# Uni Class Payload -> Py Protocol Data (for interfaces)
def class_payload_to_protocol_data(
    data: ClassPayload,
    class_uid: DefnUID,
    ctx: dict[DefnUID, DefUniMsg],
    synthetic_defs: list[DefUniMsg] | None = None,
) -> ClassDataMsg:
    # Get the metaclass (type) as a system resource
    type_sid = LRID_TO_SRID.get(id(type), 0)
    cls_msg = SystemResourceMsg(sid=type_sid)

    uni_bases = data.bases if data.bases is not UNSET else ()

    cleaned_uni_bases = list()
    for base in uni_bases:
        base_def = get_def_or_sysref_from_ctx(ctx, base)
        if (
            isinstance(base_def, ClassDefUniMsg)
            and base_def.payload.instance_of_generic is not UNSET
            and base_def.payload.instance_of_generic.resource in EXCLUDE_FROM_BASES
        ):
            continue
        cleaned_uni_bases.append(base)

    # Check if any base is/extends an interface (will become a Protocol)
    # If so, we shouldn't add Protocol again (would cause MRO conflict)
    has_interface_base = any(uni_inherits_from_interface(b, ctx) for b in cleaned_uni_bases)

    # Convert interface bases to Python terms
    interface_bases = tuple(uni_to_py_term(b, ctx) for b in cleaned_uni_bases)

    # Add Protocol as the first base class, only when no base is/extends an interface
    if not has_interface_base:
        protocol_sid = LRID_TO_SRID.get(id(TypingProtocol), 0)
        protocol_msg = SystemResourceMsg(sid=protocol_sid)
        interface_bases = (protocol_msg,) + interface_bases

    # Synthesize __init__ from constructor arguments
    new_synthetic_init = add_init_method(data, class_uid, ctx)
    if synthetic_defs is not None:
        synthetic_defs.extend(new_synthetic_init)

    # Add synthetic dict protocol methods if the interface has an index signature
    new_synthetic = add_dict_protocol_methods(data, class_uid, ctx)
    if synthetic_defs is not None:
        synthetic_defs.extend(new_synthetic)

    # methods: mapping name -> resource (function def/ref)
    methods: dict[str, tuple[str, ResourceMsg]] = {}
    if data.methods is not UNSET:
        for m in data.methods:
            name = m.name
            # skip private methods
            if m.is_private is not UNSET and m.is_private:
                continue
            fun_ref = uni_to_py_term(m.function, ctx)
            assert isinstance(fun_ref, ResourceMsg)
            methods[name] = ('instance', fun_ref)

    # annotations to fields
    annos, _instance_keys, class_keys = make_annos_and_keys(data, class_uid, ctx, synthetic_defs)

    return ClassDataMsg(  # type: ignore[call-arg]
        name=data.name,
        cls=cls_msg,
        bases=interface_bases,  # type: ignore[arg-type]
        methods=methods,
        qname=data.name,
        module="__main__",  # TODO: maybe one day "__main__" if data.module is UNSET else data.module,
        doc="" if data.doc is UNSET else data.doc,
        annos=annos,
        keys=tuple(class_keys),
    )


# Uni Fun Payload -> Py Fun Data
def function_payload_to_py_data(
    data: FunctionPayload,
    ctx: dict[DefnUID, DefUniMsg],
    class_def: ClassDefUniMsg | InterfaceUniMsg | IntersectionUniMsg | None = None,
) -> FunctionDataMsg:
    # Build argument metadata:
    #    args_to_rest_opt: name -> (is_rest, is_optional)
    args_to_rest_opt: dict[str, tuple[bool, bool]] = dict()
    if data.arguments is not UNSET:
        args_to_rest_opt = {
            arg.name: (
                arg.rest is not UNSET and arg.rest,
                (arg.optional is not UNSET and arg.optional) or arg.default is not UNSET,
            )
            for arg in data.arguments
        }

    if args_to_rest_opt:
        rest_args = tuple(k for k, (is_rest, _) in args_to_rest_opt.items() if is_rest)
        pos_star: str | None = rest_args[0] if rest_args else None
        arg_names: tuple[str, ...] = tuple(
            k for k, (is_rest, _) in args_to_rest_opt.items() if not is_rest
        )
        opt_args: tuple[str, ...] = tuple(
            k for k, (is_rest, is_opt) in args_to_rest_opt.items() if is_opt and not is_rest
        )
    else:
        pos_star = None
        arg_names = ()
        opt_args = ()

    # annotations
    annos: dict[str, TermMsg] = {}
    if data.arguments is not UNSET:
        for arg in data.arguments:
            try:
                annos[arg.name] = uni_to_py_term(arg.type, ctx)
            except:
                # TODO: should log / notify about this
                # best-effort: skip unsupported annotations
                pass

    # Default arguments
    defaults: dict[str, TermMsg] = {}
    if data.arguments is not UNSET:
        for arg in data.arguments:
            if arg.default is not UNSET:
                try:
                    defaults[arg.name] = uni_to_py_term(arg.default, ctx)
                except:
                    # TODO: should log / notify about this
                    pass

    # Return type
    if data.returnType is not UNSET:
        annos['return'] = uni_to_py_term(data.returnType, ctx)

    # If this is a method attached to a class, compute qname and prepend self
    qname = data.name
    if class_def is not None:
        # IntersectionPayload doesn't have a name field, so we just use the method name
        if isinstance(class_def, IntersectionUniMsg):
            qname = data.name
        else:
            qname = f"{class_def.payload.name}.{data.name}"
        # Prepend 'self' to args if not already there
        if not arg_names or arg_names[0] != 'self':
            arg_names = ('self',) + tuple(arg_names)
        # Annotate self with the class type
        self_rid = resourceUIDToGrid(class_def.uid)
        # HACK: don't annotate self for now ... it errors due to recursion
        # annos['self'] = uni_to_py_term(
        #     class_def, ctx
        # )  # TODO: could turn this into a string annotation

    return FunctionDataMsg(  # type: ignore[call-arg]
        name=data.name,
        qname=qname,
        module="__main__",  # TODO: maybe one day "__main__" if data.module is UNSET else data.module,
        doc=None if data.doc is UNSET else data.doc,
        keys=(),
        args=arg_names,
        opt_args=opt_args,
        pos_star=pos_star,
        annos=annos,
        defaults=defaults,
    )


# Uni Object Payload -> Py Object Data
def object_payload_to_py_data(
    payload: ObjectPayload, ctx: dict[DefnUID, DefUniMsg]
) -> ObjectDataMsg | FutureDataMsg:
    if payload.cls is UNSET:
        cls_def = RefUniMsg(uid=DefnUID(world='client', resource=BUILTIN_UNI_IDS["Any"]))
    else:
        cls_def = get_def_or_sysref_from_ctx(ctx, payload.cls)
        if cls_def is None:
            raise ValueError(f"TranscoderError: class not found in context {payload.cls}")

    if isinstance(cls_def, DefinitionMsg) and isinstance(cls_def.data, ForwardRefDataMsg):
        return ForwardRefDataMsg(name=cls_def.data.name)

    ## CASE 1: union-typed objects have their own code path
    if isinstance(cls_def, UnionUniMsg):
        # For union-typed objects, match runtime keys to a specific disjunct. If we can't find
        # a matching disjunct, then ... uh, well it's better than not having the working paths.
        runtime_keys = set(payload.keys) if payload.keys is not UNSET else set()

        matched_disjunct = None
        for disjunct_ref in cls_def.payload.classes:
            disjunct_def = get_def_or_sysref_from_ctx(ctx, disjunct_ref)
            if disjunct_def is None:
                # Ignore disjuncts that are not in the context... (e.g. primitive values)
                continue

            def get_expected_keys(cls_payload: ClassPayload, expected_keys: set[str]):
                if cls_payload.methods is not UNSET:
                    expected_keys.update(
                        {
                            m.name for m in cls_payload.methods if m.function.uid.resource > 0
                        }  # we want to exclude synthetic methods ... > 0 !
                    )
                if cls_payload.fields is not UNSET:
                    expected_keys.update(
                        {
                            f.name for f in cls_payload.fields if f.is_optional is not True
                        }  # we want to exclude optional fields
                    )

            expected_keys = set()
            if isinstance(disjunct_def, InterfaceUniMsg):
                get_expected_keys(disjunct_def.payload, expected_keys)
            elif isinstance(disjunct_def, ClassDefUniMsg):
                get_expected_keys(disjunct_def.payload, expected_keys)
            if expected_keys.issubset(runtime_keys):
                matched_disjunct = disjunct_ref

        if matched_disjunct is not None:
            cls_ref_py = uni_to_py_term(matched_disjunct, ctx)
            matched_def = ctx[matched_disjunct.uid]

            if isinstance(matched_def, InterfaceUniMsg):
                if matched_def.payload.methods is not UNSET:
                    keys = tuple(
                        m.name
                        for m in matched_def.payload.methods
                        if m.is_private is not UNSET and not m.is_private
                    )
                else:
                    keys = ()
            elif isinstance(matched_def, ClassDefUniMsg):
                if matched_def.payload.fields is not UNSET:
                    keys = tuple(
                        f.name
                        for f in matched_def.payload.fields
                        if f.is_private is not UNSET and not f.is_private
                    )
                else:
                    keys = ()
            else:
                keys = tuple(payload.keys) if payload.keys is not UNSET else ()
        else:
            raise ValueError(
                "Error during union-typed object disambiguation: did not find a matching disjunct"
            )

        return ObjectDataMsg(cls=cls_ref_py, keys=uniq(keys))  # type: ignore[call-arg]

    ## CASE 2: Objects of system classes
    elif isinstance(cls_def, RefUniMsg) and cls_def.uid.resource < 0:
        cls_ref_py = uni_to_py_term(cls_def, ctx)
        keys = tuple(uniq(payload.keys)) if payload.keys is not UNSET else ()
        return ObjectDataMsg(cls=cls_ref_py, keys=keys)  # type: ignore[call-arg]

    ## Non-system and non-union objects
    else:
        ## Expected class types:
        # Objects can have ClassDefUniMsg (concrete class), InterfaceUniMsg (Protocol),
        # or IntersectionUniMsg (Protocol with multiple inheritance) as their class
        assert isinstance(cls_def, (ClassDefUniMsg, InterfaceUniMsg, IntersectionUniMsg)), (
            f"Expected ClassDefUniMsg, InterfaceUniMsg, or IntersectionUniMsg for obj.cls, got {cls_def}"
        )

        ## key extraction logic
        # Extract keys based on the class type
        if isinstance(cls_def, IntersectionUniMsg):
            # For intersections, collect keys from all component types
            keys_list = []
            for component_ref in cls_def.payload.classes:
                component_def = get_def_or_sysref_from_ctx(ctx, component_ref)
                if (
                    component_def
                    and isinstance(component_def, (ClassDefUniMsg, InterfaceUniMsg))
                    and component_def.payload.fields is not UNSET
                ):
                    keys_list.extend(
                        f.name
                        for f in component_def.payload.fields
                        if f.is_private is not UNSET and not f.is_private
                    )  # type: ignore[attr-defined]
            keys = tuple(keys_list)
        elif hasattr(cls_def.payload, 'fields') and not isinstance(
            cls_def.payload.fields, UnsetType
        ):
            # For ClassDefUniMsg and InterfaceUniMsg with fields
            keys = tuple(
                f.name
                for f in cls_def.payload.fields
                if f.is_private is not UNSET and not f.is_private
            )
        else:
            keys = ()

        cls_ref_py = uni_to_py_term(referentialize(cls_def), ctx)

        # Add object-specific keys
        if payload.keys is not UNSET:
            keys = keys + tuple(payload.keys)

        return ObjectDataMsg(cls=cls_ref_py, keys=uniq(keys))  # type: ignore[call-arg]


def uniq(xs):
    seen = set()
    cls = xs.__class__
    return cls(x for x in xs if x not in seen and not seen.add(x))


### Helper functions


def uni_to_py_frame_request(req: RequestUniMsg) -> tuple[FrameID, FrameID, MessageID]:
    pid = req.parentFID
    fid = req.selfFID
    mid = req.requestedFID

    return pid, fid, mid


def uni_to_py_frame_response(res: ResponseUniMsg) -> tuple[FrameID, FrameID]:
    pid = res.parentFID
    fid = res.selfFID

    return pid, fid


def uni_to_ordered_py_defs(
    defs: list[DefUniMsg], ctx: dict[DefnUID, DefUniMsg]
) -> list[DefinitionMsg]:
    ordered_defs, _ = order_defs(defs)
    py_defs: list[DefinitionMsg] = []
    for d in ordered_defs:
        try:
            # NOTE: this is the runtime code path, we ensure that the synths appear before the parent
            # but this is effectively not exercised by TS runtime so is untested
            synthetic_defs: list[DefUniMsg] = []
            res = uni_to_py_def(d, ctx, None, synthetic_defs)
            for d in synthetic_defs:
                res = uni_to_py_def(d, ctx)
                py_defs.append(res)
            py_defs.append(res)
        except NotImplementedError:
            raise ValueError(f"Unsupported universal def: {d}")
    return py_defs


def make_py_call_args(
    args: list[CallArg],
    ctx: dict[DefnUID, DefUniMsg],
) -> tuple[tuple[TermMsg, ...], dict[str, TermMsg]]:
    # for now, everything is a positional argument
    # TODO: Could deduce key args from ordering in defs
    pos: list[TermMsg] = []
    key: dict[str, TermMsg] = {}
    for arg in args:
        v = arg.val
        py_v = uni_to_py_term(v, ctx)
        pos.append(py_v)
    return tuple(pos), key



import re
from itertools import chain, count, islice, repeat

__all__ = [
    'ITERATOR_FNS',
]

dct = dict(a=1, b=2, c=3)
tup = (1, 2, 3)
lst = [1, 2, 3]
st = {1, 2, 3}


def gen_fn():
    yield 1
    yield 2
    yield 3


def counter():
    i = 0

    def _next():
        nonlocal i
        i += 1
        return i

    return _next


patt = re.compile('.')

ITERATOR_FNS = (
    lambda: range(5),
    lambda: map(str, lst),
    lambda: filter(str.isupper, 'aBcDeF'),
    lambda: zip(tup, lst),
    lambda: iter(tup),
    lambda: iter(lst),
    lambda: iter(st),
    lambda: dct.items(),
    lambda: dct.keys(),
    lambda: dct.values(),
    lambda: chain(lst, lst, lst),
    lambda: islice(lst, 2),
    lambda: count(0, 3),
    lambda: repeat(9),
    lambda: (str(i) for i in lst),
    lambda: gen_fn(),
    lambda: iter(counter(), 5),
)



import re
from pathlib import Path
from typing import Callable, Self

from ...core.strs import pathsafe_str

__all__ = [
    'FileSection',
    'FileTemplate',
]

###############################################################################


class FileSection(Path):
    base: Path
    elem: str
    _key: str
    _pat: str
    _sin: Callable[[str], str]
    _mul: Callable[[str], str]

    def __init__(self, path: str | Path, elem: str) -> None:
        super().__init__(path)
        self.base = Path(path)
        self.elem = elem
        key = f'⟨{elem}⟩'
        pad = key.ljust(32)
        lhs = re.escape(key)
        self._key = key = f'⟨{elem}⟩'
        self._pat = f'{lhs}\\s*⟪{VAL}⟫\n'
        self._sin = f'{pad}⟪{{}}⟫\n'.format
        self._mul = f'{key}\n⟪{{}}⟫\n'.format

    def __str__(self) -> str:
        line = 0
        if self.__exists():
            if txt := self.__read():
                key = self._key
                if key in txt:
                    for i, l in enumerate(txt.splitlines()):
                        if l.startswith(key):
                            line = i + 1
                            break
        file = self.base.as_posix()
        if line:
            file = f'{file}:{line}'
        return f'{file}#{self.elem}'

    def __repr__(self) -> str:
        return f'SegmentedFile({str(self)!r})'

    def __fspath__(self) -> str:
        return self.base.as_posix()

    def absolute(self) -> Self:
        if self.base.is_absolute():
            return self
        return FileSection(self.base.absolute(), self.elem)

    def exists(self, **_) -> bool:
        return self.__exists() and self._key in self.__read()

    def unlink(self, **_):
        text = self.__read()
        if self._key not in text:
            return
        if text := re.sub(self._pat, '', text):
            self.__write(text)
        else:
            self.base.unlink()

    def write_text(self, text: str, **_) -> None:
        prev = self.__read()
        new = self._mul(text) if '\n' in text else self._sin(text)
        if self._key not in prev:
            self.__write(prev + new)
        else:
            self.__write(re.sub(self._pat, new, prev))

    @property
    def parent(self) -> Path:
        return self.base.parent

    def read_text(self, **_) -> str:
        text = self.__read()
        if self._key not in text:
            raise RuntimeError(f'segment {self.elem!r} not found in {self.__base}')
        pat = self._pat
        match = re.search(pat, text)
        if match is None:
            raise RuntimeError(f'RE did not match: {pat!r}')
        return match.group(1)

    def __exists(self) -> bool:
        base = self.base
        return base.exists() and base.is_file()

    def __write(self, s: str):
        self.base.write_text(s)

    def __read(self) -> str:
        base = self.base
        if not base.exists():
            return ''
        assert base.is_file(), f'{base!r} not a file'
        return base.read_text()


VAL = r'([^⟪⟫]*)'


class FileTemplate:
    parts: list[str | Path]
    section: str | None

    def __init__(self, *parts: str | Path, section: str = None) -> None:
        *path, file = list(parts)
        path = [p.removesuffix('.py') for p in path]
        self.parts = [*path, file]
        self.section = section

    # def absolute(self) -> Path:
    #     parts = self.parts.copy()
    #     parts[0] = parts[0].absolute()
    #     return replace(self, parts=parts)

    def __str__(self) -> str:
        return '~' + '/'.join(self.parts) + f'#{self.section}' if self.section else ''

    def __repr__(self) -> str:
        return f'FileTemplate(parts={self.parts!r}, section={self.section!r})'

    def format(self, **kwargs):
        parts = self.parts
        section = self.section
        safe = {k: pathsafe_str(v) for k, v in kwargs.items()}
        parts = [p.format(**safe) if type(p) is str else p for p in parts]
        path = Path(*parts)
        if not section:
            return path
        if path.exists() and path.is_dir():
            raise ValueError('base path {path!r} already exists and is a directory')
        section = section.format(**kwargs)
        return FileSection(path, section)



# fmt: off

from dataclasses import dataclass
from .repl_alias import Options

__all__ = [
    'ReplEvaluationInfo',
]

################################################################################

@dataclass
class ReplEvaluationInfo:
    """
    Represents a summary of the contents of a `ReplEvaluationData` object.

    warp will automatically JSON-serialize instances within REPL messages
    since they operate with `fmt=JSON`.
    """

    output:             str
    out_str:            str | None

    traceback_str:      str | None
    exception_name:     str | None
    has_return_value:   bool
    has_raised_error:   bool
    has_uncaught_error: bool

    added_locals:       tuple[str, ...]
    changed_locals:     tuple[str, ...]
    removed_locals:     tuple[str, ...]

    imported:           tuple[str, ...]

    duration:           float
    metadata:           Options

    @staticmethod
    def empty() -> 'ReplEvaluationInfo':
        return ReplEvaluationInfo(
            output='', out_str=None,
            traceback_str=None,
            exception_name=None,
            has_return_value=False,
            has_raised_error=False,
            has_uncaught_error=False,
            added_locals=(),
            changed_locals=(),
            removed_locals=(),
            imported=(),
            duration=0,
            metadata={},
        )

    @property
    def has_error(self):
        return self.has_uncaught_error or self.has_raised_error

    @property
    def has_result(self):
        return self.has_return_value or self.has_raised_error

    def __debug_info_str__(self) -> str:
        strs = []
        add = strs.append
        if output := self.output:
            add(f'output=<{len(output)}>')
        if out_str := self.out_str:
            add(f'out_str=<{len(out_str)}>')
        if self.has_uncaught_error:
            add('has_uncaught_error=True')
        if self.has_raised_error:
            add('has_raised_error=True')
        if self.has_return_value:
            add('has_return_value=True')
        if name := self.exception_name:
            add(f'exception_name={name!r}')
        return ' '.join(strs)



# fmt: off

import sys
import builtins

from contextlib import contextmanager
from collections.abc import Callable
from typing import NamedTuple, TextIO

__all__ = [
    'SystemHooks',
]

################################################################################

type PrintFn = Callable[..., None]

class SystemHooks(NamedTuple):
    """
    Represents the changes of global Python state that must be made during
    a REPL execution.

    Make a desired state with `state = SystemHooks.make(...)`.

    Temporarily apply it with `with state.applied(): ...`

    Permanently apply it with `state.set()`.

    Get the current system state with `SystemHooks.get()`.
    """

    stdin:           TextIO
    stdout:          TextIO
    stderr:          TextIO
    print_fn:        PrintFn
    recursion_limit: int

    @staticmethod
    def make(
            stdin: TextIO,
            stdout: TextIO,
            stderr: TextIO, *,
            print_fn: PrintFn = builtins.print,
            recursion_limit: int = 0) -> 'SystemHooks':

        return SystemHooks(
            stdin, stdout, stderr,
            print_fn,
            recursion_limit,
        )

    def set(self):
        sys.stdin, sys.stdout, sys.stderr, builtins.print, recursion_limit = self
        sys.setrecursionlimit(recursion_limit) if recursion_limit else None

    @staticmethod
    def get() -> 'SystemHooks':
        rec_limit = sys.getrecursionlimit()
        return SystemHooks(
            sys.stdin, sys.stdout, sys.stderr,
            builtins.print,
            rec_limit,
        )

    @contextmanager
    def applied(self):
        old = SystemHooks.get()
        self.set()
        yield
        old.set()



# fmt: off

from .__ import *


__all__ = [
    'Request',
    'ResultCoro',
]


type ResultCoro = Coroutine[None, None, Result]

################################################################################

# NOTE: if you are curious about the fact that this class hierarchy
# seemingly duplicates the message hierarchy in RequestMsg,
# see HOUSEKEEPING.MD

class Request(ABC):
    """
    ABC for a fully decoded request.

    Instances be executed via `.execute` or `.execute_async` methods.

    Instances can be encoded to a corresponding `ResourceMsg` via `.encode`.

    If the class has slot `_async`, the property `.as_async` will declare the
    intent for the request to be executed async -- the Frame will call
    `.execute_async()` in that case.
    """

    __slots__ = __match_args__ = ()

    def __repr__(self):
        f_head = type(self).__name__
        f_args = ', '.join(self.__fmt_args__())
        return f'{f_head}({f_args})'

    def __fmt_args__(self):
        for slot in self.__slots__:
            value = getattr(self, slot, FIELD_ABSENT)
            if value is not FIELD_ABSENT:
                yield f_object_id(value)

    def pprint(self, err: bool = False):
        from ...core.fmt import f_slot_obj
        P.rprint(f_slot_obj(self), err=err)

    __short_str__ = __str__ = __repr__

    ############################################################################

    @abstractmethod
    def execute(self) -> Result:
        """Immediately execute a decoded remote request."""

    @abstractmethod
    def execute_async(self) -> ResultCoro:
        """
        Return a coroutine that executes a decoded remote request, (which is assumed
        to provide an Awaitable), awaits this, and provides the Result to result_fn."""

    @abstractmethod
    def encode(self, codec: 'EncoderP') -> 'RequestMsg':
        """Encode a locally expressed request into a wire message."""

    def async_name(self) -> str:
        """Name to give to an `execute_async` coroutine, to aid debugging."""
        return 'asynchronous operation'

    ############################################################################

    @property
    def is_async(self) -> bool:
        """True if the request is intended to be executed asynchronously."""
        return getattr(self, 'async_mode', None) in ('coro', 'future')

    @property
    def is_sync(self) -> bool:
        """True if the request is intended to be executed synchronously."""
        return getattr(self, 'async_mode', None) in (None, 'sync')

    @property
    def as_coro(self) -> bool:
        """True if the request is intended to be executed in a coroutine."""
        return getattr(self, 'async_mode', None) == 'coro'

    @property
    def as_future(self) -> bool:
        """True if the request is intended to be executed in a future."""
        return getattr(self, 'async_mode', None) == 'future'

    # --------------------------------------------------------------------------

    def set_async_mode(self, async_mode: AsyncMode) -> Self:
        """
        If not None, set this request to be executed asynchronously in one of three ways,
        'sync', 'coro', and 'future'. Locally, these behave as follows:

        * 'sync':
            block until result is returned
        * 'coro':
            immediately return a coroutine object that, when scheduled will:
            1) immediately send a FramedRequestMsg
            2) wait for and then decode the FramedResponseMsg
        * 'future':
            1) schedule a FramedRequestMsg to be sent *at some point*
            2) immediately return a Future (future_id=message_id)
            This future will be set when the FramedResponseMsg is sent

        Remotely, all three of these assume the underlying function being called returns
        an Awaitable object, and will await this result and return it via a FramedResponseMsg.

        The 'future' behavior effectively piggybacks on the logic for handling ordinary, non-asynchronous
        requests. These requests create futures for FramedResponseMsgs, from these we create a second
        future, and attach a done_callback to the first future that merely sets this second future to
        the result obtained from decoding the first future.
        """
        if async_mode:
            setattr(self, 'async_mode', async_mode)
        return self



# fmt: off

"""
This file provides a subclass of Future that accepts per-instance hooks that
are run before the default implementations of `await future`, and replace
the default implementations of `future.set_result()`, `future.set_exception()`,
and `future.cancel()`.

The await hook is needed for warp's `AgentWorld` to implement virtual/mirrored
futures correctly, since `await future` must ensure that a background
`futures_task` is started/running.

The other three hooks are used to implement *all* virtual future functionality:
1) `set_result` and `set_exception` should issue CompleteFuture RPC messages
2) `future.cancel()` should issue a CancelFuture RPC message
3) the owning event loop being shut down should NOT issue a CancelFuture RPC message

HookableFuture spoofs the name and module of `asyncio` since it is
indistinguishable in behavior.
"""

import asyncio
from asyncio import InvalidStateError

from typing import Any, TYPE_CHECKING
from collections.abc import Callable


__all__ = [
    'new_hookable_future',
    'HookableFuture',
    'HookMethod',
]


################################################################################

type HookMethod = Callable[['HookableFuture', Any], None]

class HookableFuture(asyncio.Future):

    if TYPE_CHECKING:
        __await_hook__:         HookMethod
        __cancel_hook__:        HookMethod
        __set_result_hook__:    HookMethod
        __set_exception_hook__: HookMethod
        __was_gathered_hook__:  HookMethod

    def __await__(self):
        if not self.done():
            # NOTE: this is used only in AgentWorld to start the futures_task
            if callable(hook := getattr(self, AWAIT_HOOK, None)):
                hook(self, None)
        return super().__await__()

    def cancel(self, msg: Any | None = None) -> bool:
        if self.done():
            return False
        if callable(hook := getattr(self, CANCEL_HOOK, None)):
            # if a hook is present, defer to its implementation
            result = hook(self, msg)
            if result is True and self.done():
                return result
        return super().cancel(msg)

    def set_result(self, result: Any) -> None:
        if self.done():
            # raise more intelligible exception than the default one
            raise InvalidStateError("cannot set future result; future already completed")
        if callable(hook := getattr(self, SET_RESULT_HOOK, None)):
            # if a hook is present, defer to its implementation
            hook(self, result)
            if self.done():
                # if the hook completed the future, don't set the result again
                return
        super().set_result(result)

    def set_exception(self, exception: type | BaseException) -> None:
        if self.done():
            # raise more intelligible exception than the default one
            raise InvalidStateError("cannot set future exception; future already completed")
        if callable(hook := getattr(self, SET_EXCEPTION_HOOK, None)):
            # if a hook is present, defer to its implementation
            hook(self, exception)
            if self.done():
                # if the hook completed the future, don't set the exception again
                return
        super().set_exception(exception)

    def add_done_callback(self, fn, /, *, context=None) -> None:
        if not self.done():
            if callable(hook := getattr(self, WAS_GATHERED_HOOK, None)):
                qualname = getattr(fn, '__qualname__', None)
                if qualname == 'gather.<locals>._done_callback':
                    hook(self, fn)
        super().add_done_callback(fn, context=context)

    def ___del_hooks___(self):
        if hasattr(self, AWAIT_HOOK):
            delattr(self, AWAIT_HOOK)
        if hasattr(self, CANCEL_HOOK):
            delattr(self, CANCEL_HOOK)
        if hasattr(self, SET_RESULT_HOOK):
            delattr(self, SET_RESULT_HOOK)
        if hasattr(self, SET_EXCEPTION_HOOK):
            delattr(self, SET_EXCEPTION_HOOK)
        if hasattr(self, WAS_GATHERED_HOOK):
            delattr(self, WAS_GATHERED_HOOK)

    def ___set_hooks___(self, *,
                        await_hook: HookMethod | None = None,
                        cancel_hook: HookMethod | None = None,
                        set_result_hook: HookMethod | None = None,
                        set_exception_hook: HookMethod | None = None,
                        was_gathered_hook: HookMethod | None = None) -> None:
        if await_hook:
            setattr(self, AWAIT_HOOK, await_hook)
        if cancel_hook:
            setattr(self, CANCEL_HOOK, cancel_hook)
        if set_result_hook:
            setattr(self, SET_RESULT_HOOK, set_result_hook)
        if set_exception_hook:
            setattr(self, SET_EXCEPTION_HOOK, set_exception_hook)
        if was_gathered_hook:
            setattr(self, WAS_GATHERED_HOOK, was_gathered_hook)


HookableFuture.__name__ = HookableFuture.__qualname__ = 'Future'
HookableFuture.__module__ = 'asyncio'

AWAIT_HOOK: str         = '___await_hook___'
CANCEL_HOOK: str        = '___cancel_hook___'
SET_RESULT_HOOK: str    = '___set_result_hook___'
SET_EXCEPTION_HOOK: str = '___set_exception_hook___'
WAS_GATHERED_HOOK: str  = '___was_gathered_hook___'

################################################################################

def new_hookable_future(loop: asyncio.AbstractEventLoop) -> HookableFuture:
    return HookableFuture(loop=loop)



# fmt: off

# this file is pretty much imported everywhere else and sets up extremely common shared imports,
# to avoid endless ruff churn and inconvenience in having to import things per-file

import collections.abc as A
import operator as O
import typing as T
from abc import ABC, abstractmethod

from collections.abc import Callable, Coroutine, Awaitable, Iterable as Iter, Sequence as Seq

from functools import partial
from types import ModuleType, FunctionType, CoroutineType
from typing import Any, ClassVar, Generic, Literal, NoReturn, Protocol, Self, TypeGuard
from typing import cast, overload, TYPE_CHECKING

from ..core import type as S
from ..core import debug as D
from ..core import fmt as F
from ..core import print as P

from ..core.make import *
from ..core.result import *
from ..core.sentinels import *
from ..core.log import LogFlag, LogContext, should_log_cls

from . import exceptions as E
from . import flags

from .kinds import *
from .fmt import *
from .alias import *
from .predicates import *


__all__ = [
    # from system modules
    'partial',
    'ModuleType', 'FunctionType', 'CoroutineType',
    'ABC', 'abstractmethod',
    'Callable', 'Coroutine', 'Awaitable', 'Iter', 'Seq',
    'Any', 'Generic', 'TypeGuard', 'Self', 'ClassVar', 'Protocol',
    'Literal', 'NoReturn', 'cast', 'overload', 'TYPE_CHECKING',

    # aliases for common modules
    'T', 'S', 'A', 'O', 'D', 'P', 'F', 'E',

    'flags',

    # from `core.log`
    'LogFlag',
    'LogContext',
    'should_log_cls',

    # from `core.sentinels`
    'ArgDefault',  'FieldAbsent',  'Pending', 'Closed', 'Canceled',
    'ARG_DEFAULT', 'FIELD_ABSENT', 'PENDING', 'CLOSED', 'CANCELED',
    'Sentinel', 'is_sentinel',

    # from `core.result`
    'Result', 'OK_RESULT',

    # from `core.make`
    'mklist', 'mkdict', 'mkset',

    # from `warpc.alias`
    'Tup', 'Rec', 'Fn', 'SendBytes', 'EncodeFmt', 'DecodeFmt', 'MethodKind', 'AsyncMode', 'VarScope',
    'optstr', 'strtup', 'strlist', 'typetup', 'record',
    'Name', 'Ptr', 'ID',
    'WorldID', 'LocalRID', 'GlobalRID', 'ResourceUID',
    'FrameKey', 'SystemRID', 'FrameID', 'MessageID', 'FutureID', 'Options',

    # from `warpc.predicates`
    'is_bool',
    'is_str',
    'is_strtup',
    'is_strlist',
    'is_optstr',
    'is_tup',
    'is_list',
    'is_rec',
    'is_bytes',

    # from `warpc.kinds`
    'Kind',
    'TermT',
    'ValueT', 'AtomT', 'NumberT', 'StrLikeT', 'SingletonT', 'ContainerT', 'SequenceT', 'MappingT',
    'ResourceT', 'ObjectT', 'ClassT', 'FunctionT', 'CoroutineT', 'ModuleT', 'TypeT', 'IteratorT',
    'FutureT',
    'ResourcesRecordT', 'MethodsT', 'ClassesTupleT',
    'ArgsT', 'KwargsT', 'AnnotationsT', 'AttributesT',

    # from `warpc.fmt`
    'f_id', 'f_grid', 'f_fkey',
    'f_slot', 'f_object', 'f_object_id',
]



# fmt: off

from typing import Any
from collections.abc import Callable
from types import NoneType

from ..core.fmt import f_object, f_object_id
from .alias import *

__all__ = [
    'f_id',
    'f_grid',
    'f_fkey',
    'f_slot',
    'f_object',
    'f_object_id',
]

################################################################################

def f_id(i: int) -> str:
    if type(i) is int:
        #return f'{i:x}' if i >= 0 else f'-{-i:x}'
        return str(i)
    return '<!id>'

def f_grid(grid: GlobalRID) -> str:
    if type(grid) is tuple and len(grid) == 3:
        wid, fid, rid = grid
        # return f'{wid:x}:{fid:x}:{rid:x}'
        return f'{wid}:{fid}:{rid}'
    return '<!grid>'

def f_fkey(fkey: FrameKey) -> str:
    if type(fkey) is tuple and len(fkey) == 2:
        wid, fid = fkey
        return f'{wid:x}:{fid:x}'
    return '<!fkey>'

def f_slot(key: str, val: Any, fn: Callable[..., str] = f_object_id) -> str:
    match val:
        case None:
            return 'None'
        case type():
            return val.__name__
        case str() | bytes() if len(val) < 12:
            return repr(val)
        case tuple():
            if key == 'grid':
                return f_grid(val)
            if not val:
                return '()'
            if len(val) == 1:
                return '(' + f_slot('', val[0], fn) + ',)'
            if _is_short(val):
                return '(' + ', '.join(f_slot('', v, fn) for v in val) + ')'
            return '(' + f'{len(val)}…' + ')'
        case list():
            if not val:
                return '[]'
            if _is_short(val):
                return '[' + ', '.join(f_slot('', v, fn) for v in val) + ']'
            return '[' + f'{len(val)}…' + ']'
        case dict():
            if not val:
                return '{}'
            return '{' + f'{len(val)}…' + '}'
        case int() | float() | bool():
            if key.endswith('id'):
                return f_id(val)
            return str(val)
        case _:
            return fn(val)

SHORT = int, bool, str, type, NoneType

def _is_short(seq) -> bool:
    return len(seq) <= 4 and all(type(v) in SHORT for v in seq)



# ruff: noqa
# fmt: off

import pytest

from agentica_internal.testing.warpc.virtual_builtin_tests import verify_virtual_queue


@pytest.mark.asyncio
async def test_virtual_queue():
    await verify_virtual_queue()



from typing import overload


@overload
def foo(x: int) -> int:
    """i love ints"""
    return x


@overload
def foo(x: str) -> str:
    """i love strings"""
    return x


def foo(x: int | str) -> int | str:
    """i love ints and strings"""
    return x



import json

from httpx import Request, Response

from .base import AgenticaError

__all__ = [
    'ServerError',
    'GenerationError',
    'UsageError',
    'MaxTokensError',
    'MaxRoundsError',
    'ContentFilteringError',
    'InferenceError',
    'APIConnectionError',
    'APITimeoutError',
    'RateLimitError',
    'BadRequestError',
    'UnauthorizedError',
    'PermissionDeniedError',
    'InsufficientCreditsError',
    'NotFoundError',
    'ConflictError',
    'UnprocessableEntityError',
    'RequestTooLargeError',
    'ServiceUnavailableError',
    'OverloadedError',
    'DeadlineExceededError',
    'InternalServerError',
]


class ServerError(AgenticaError):
    """Base class for exceptions during remote operations."""

    http_status_code: int | None = None

    def __init__(self, message: str, http_status_code: int | None = None):
        super().__init__(message)
        self.http_status_code = http_status_code

    def __str__(self) -> str:
        s = super().__str__()
        if self.http_status_code is not None:
            s += f" (HTTP {self.http_status_code})"
        return s


# === Base Exceptions ===


class GenerationError(ServerError):
    """Base class for exceptions during agent generation."""

    def __init__(self, message: str):
        super().__init__(message)


# === Generation Exceptions ===


class UsageError(GenerationError): ...


class MaxTokensError(UsageError):
    """Max tokens error."""

    def __init__(self, max_tokens: int | str):
        super().__init__(
            message=(
                f"The maximum number of tokens ({max_tokens}) has been reached."
                if type(max_tokens) is int
                else str(max_tokens)
            )
        )


class MaxRoundsError(UsageError):
    """Max rounds error."""

    def __init__(self, max_rounds: int | str):
        super().__init__(
            message=(
                f"The maximum number of rounds of inference ({max_rounds}) has been reached."
                if type(max_rounds) is int
                else str(max_rounds)
            )
        )


class ContentFilteringError(GenerationError):
    """Content filtering error."""

    def __init__(self):
        super().__init__(f"The previously generated content has been filtered.")


class InferenceError(GenerationError):
    """Base class for exceptions during inference, mainly HTTP errors."""

    request: Request
    response: Response | None

    def __init__(
        self,
        request: Request,
        response: Response | None = None,
        prefix: str = '',
        message: str = 'Unknown inference error.',
    ):
        if isinstance(request, str):
            # Bad usages of this exception are extremely common.
            message = request

        self.request = request
        self.response = response
        self.prefix = prefix

        if response is not None:
            if response.is_closed and not response.is_stream_consumed:
                body = None
                message = f"Error code: {response.status_code}"
            else:
                err_text = response.text.strip()
                body = err_text

                try:
                    body = json.loads(err_text)
                    message = f"Error code: {response.status_code} - {body}"
                except Exception:
                    message = err_text or f"Error code: {response.status_code}"
        if prefix and prefix.strip():
            message = f"{prefix}. {message}"
        super().__init__(message)


# === Inference errors ===


class APIConnectionError(InferenceError):
    """API connection error."""


class APITimeoutError(InferenceError):
    """API timeout error."""


# HTTP status errors


class RateLimitError(InferenceError):
    """Rate limit error."""


class BadRequestError(InferenceError):
    """Bad request error."""


class UnauthorizedError(InferenceError):
    """Unauthorized error."""


class PermissionDeniedError(InferenceError):
    """Permission denied error."""


class InsufficientCreditsError(InferenceError):
    """Insufficient credits error."""


class NotFoundError(InferenceError):
    """Not found error."""


class ConflictError(InferenceError):
    """Conflict error."""


class UnprocessableEntityError(InferenceError):
    """Unprocessable entity error."""


class RequestTooLargeError(InferenceError):
    """Request too large error."""


class ServiceUnavailableError(InferenceError):
    """Service unavailable error."""


class OverloadedError(InferenceError):
    """Overloaded error."""


class DeadlineExceededError(InferenceError):
    """Deadline exceeded error."""


class InternalServerError(InferenceError):
    """Internal server error."""



from agentica_internal.testing.cpython.shed_tests import *


def test_builtins_shed():
    verify_builtins_shed()


def test_math_shed():
    verify_math_shed()


def test_not_in_shed():
    verify_not_in_shed()



# fmt: off

import asyncio
from enum import StrEnum
from collections.abc import Callable, Coroutine, Iterator
from types import (
    FunctionType,
    CoroutineType,
    EllipsisType,
    GeneratorType,
    AsyncGeneratorType,
    ModuleType,
    NoneType,
    NotImplementedType,
)
from typing import Any, TypeGuard  # type: ignore

from ..cpython.alias import CALLABLES as FUNC_TYPES, PROPERTIES as PROP_TYPES
from ..cpython.classes.anno import ANNOS, AnnoT

from .alias import Rec, Tup, MethodKind


__all__ = [
    'Kind',
    'TermT',
    'TypeT',
    'ValueT',
        'AtomT',
            'NumberT',
            'StrLikeT',
            'SingletonT',
        'ContainerT',
            'SequenceT',
            'MappingT',
    'ResourceT',
        'ObjectT',
        'ClassT',
        'FunctionT',
        'ModuleT',
        'CoroutineT',
        'IteratorT',
        'FutureT',
    'ArgsT',
    'KwargsT',
    'AnnotationsT',
    'AttributesT',
    'ClassesTupleT',
    'ResourcesRecordT',
    'MethodT',
    'MethodsT',
    'is_atom_t',
    'is_object_t',
    'is_class_t',
    'is_function_t',
    'is_method_t',
    'is_property_t',
    'is_coroutine_t',
    'is_module_t',
    'is_type_t',
    'is_generator_t',
    'is_iterator_t',
    'is_future_t',
    'pack_method',
    'unpack_method',
]


################################################################################

class Kind(StrEnum):
    Object    = 'object'
    Class     = 'class'
    Coroutine = 'coroutine'
    Type      = 'type'
    Function  = 'function'
    Module    = 'module'
    Iterator  = 'iterator'
    Generator = 'generator'
    Future    = 'future'
    Enum      = 'enum'
    Unknown   = 'unknown'

################################################################################

type TermT      = Any
type NumberT    = bool | int | float
type StrLikeT   = str | bytes
type SingletonT = NoneType | NotImplementedType | EllipsisType
type AtomT      = NumberT | StrLikeT | SingletonT
type SequenceT  = list | dict | set | frozenset
type MappingT   = dict
type ContainerT = SequenceT | MappingT
type ValueT     = AtomT | ContainerT
type ObjectT    = object
type TypeT      = AnnoT
type ClassT     = type
type FunctionT  = Callable[..., Any]
type MethodT    = FunctionType | staticmethod | classmethod
type CoroutineT = CoroutineType
type ModuleT    = ModuleType
type IteratorT  = GeneratorType | Iterator
type FutureT    = asyncio.Future
type ResourceT  = ObjectT | ClassT | FunctionT | ModuleT | IteratorT | FutureT

################################################################################

# these make it easy to recognize certain patterns for codec compilation
type ArgsT            = Tup[TermT]
type KwargsT          = Rec[TermT]
type AnnotationsT     = Rec[TypeT]
type AttributesT      = Rec[TermT]
type ClassesTupleT    = Tup[ClassT]
type ResourcesRecordT = Rec[ResourceT]
type MethodsT         = Rec[MethodT]

################################################################################

TYPE_TYPES = ANNOS

GENERATOR_TYPES = AsyncGeneratorType, GeneratorType,

ATOM_TYPES = bool, int, float, str, bytes, None, NotImplementedType, EllipsisType

NON_OBJECT_TYPES = FUNC_TYPES + TYPE_TYPES + GENERATOR_TYPES + (type, ModuleType, CoroutineType)

METHOD_TYPES = FunctionType, staticmethod, classmethod

def is_atom_t(arg: Any) -> TypeGuard[AtomT]:
    return type(arg) in ATOM_TYPES

def is_object_t(arg: Any) -> TypeGuard[ObjectT]:
    return not issubclass(type(arg), NON_OBJECT_TYPES)

def is_class_t(arg: Any) -> TypeGuard[ClassT]:
    return type(arg) is type or isinstance(arg, type)

def is_coroutine_t(arg: Any) -> TypeGuard[CoroutineT]:
    return type(arg) is CoroutineType or isinstance(arg, Coroutine)

def is_function_t(arg: Any) -> TypeGuard[FunctionT]:
    return issubclass(type(arg), FUNC_TYPES)

def is_method_t(arg: Any) -> TypeGuard[MethodT]:
    return type(arg) in METHOD_TYPES

def is_property_t(arg: Any) -> bool:
    return type(arg) in PROP_TYPES

def is_module_t(arg: Any) -> TypeGuard[ModuleT]:
    return issubclass(type(arg), ModuleType)

def is_type_t(arg: Any) -> TypeGuard[TypeT]:
    return issubclass(type(arg), TYPE_TYPES)

def is_generator_t(arg: Any) -> TypeGuard[IteratorT]:
    return type(arg) in GENERATOR_TYPES

def is_iterator_t(arg: Any) -> TypeGuard[IteratorT]:
    return type(arg) in GENERATOR_TYPES or isinstance(arg, Iterator)

def is_future_t(arg: Any) -> TypeGuard[FutureT]:
    return isinstance(arg, asyncio.Future)

################################################################################

def pack_method(kind: MethodKind, func: FunctionT) -> MethodT:
    if kind == 'class':
        return classmethod(func)  # type: ignore
    elif kind == 'static':
        return staticmethod(func)
    elif kind == 'instance':
        return func  # type: ignore
    else:
        raise ValueError(kind)

def unpack_method(method: MethodT) -> tuple[MethodKind, FunctionT]:
    if isinstance(method, classmethod):
        return 'class', method.__func__
    elif isinstance(method, staticmethod):
        return 'static', method.__func__
    else:
        return 'instance', method



# ruff: noqa
# fmt: off

import pytest
from agentica_internal.testing.warpc.async_coro_tests import *
from agentica_internal.testing.warpc.async_future_tests import *
from agentica_internal.testing.warpc.virtual_future_tests import *


@pytest.mark.asyncio
async def test_coro_function():
    await verify_coro_function()


@pytest.mark.asyncio
async def test_coro_functions_run_concurrently():
    await verify_coro_functions_run_concurrently()


@pytest.mark.asyncio
async def test_future_function_ok():
    await verify_future_function_ok()


@pytest.mark.asyncio
async def test_future_function_error():
    await verify_future_function_error()


@pytest.mark.asyncio
async def test_virtual_future_properties():
    await verify_virtual_future_properties()


@pytest.mark.asyncio
async def test_virtual_cancellation():
    await verify_virtual_cancellation()


@pytest.mark.asyncio
async def test_virtual_completion():
    await verify_virtual_completion()


@pytest.mark.asyncio
async def test_real_cancellation():
    await verify_real_cancellation()


@pytest.mark.asyncio
async def test_real_completion():
    await verify_real_completion()



from . import base


class ConnectionError(base.AgenticaError):
    """Base class for connection errors."""

    ...


class WebSocketConnectionError(ConnectionError):
    """WebSocket connection error."""

    ...


class WebSocketTimeoutError(ConnectionError):
    """WebSocket timeout error."""

    ...


class SDKUnsupportedError(base.AgenticaError):
    """Raised when the SDK version is no longer supported by the server."""

    ...


class ClientServerOutOfSyncError(base.AgenticaError):
    """Raised when the client and server are out of sync and cannot be recovered."""

    ...



from collections import defaultdict
from types import EllipsisType, NoneType, NotImplementedType

__all__ = [
    'get_type_slots',
    'get_type_slots_all',
    'drop_system_slots',
    'NO_SLOT_TYPES',
    'SYSTEM_SLOTS',
]

NO_SLOT_TYPES = (
    type,
    object,
    int,
    float,
    complex,
    str,
    list,
    set,
    dict,
    tuple,
    defaultdict,
    frozenset,
    bytes,
    bytearray,
    NoneType,
    EllipsisType,
    NotImplementedType,
    type,
)

SYSTEM_SLOTS = (
    '__hashinfo__',
    '__uuid__',
    '__weakref__',
)


def drop_system_slots(slots: tuple[str, ...]) -> tuple[str, ...]:
    return tuple(s for s in slots if s not in SYSTEM_SLOTS)


def get_type_slots(cls: type) -> tuple[str, ...]:
    if cls in NO_SLOT_TYPES:
        return ()
    slots = getattr(cls, '__slots__', ())
    if not slots:
        return slots
    if isinstance(slots, str):
        return (slots,)
    if not isinstance(slots, tuple):
        slots = tuple(slots)
    assert all(type(s) is str for s in slots)
    return slots


def get_type_slots_all(cls: type) -> tuple[str, ...]:
    if cls in NO_SLOT_TYPES:
        return ()
    if '__slotnames__' in cls.__dict__:
        slotnames = getattr(cls, '__slotnames__')
        return tuple(slotnames)
    res = []
    for sup in reversed(cls.__mro__):
        sup_slots = get_type_slots(sup)
        res.extend(sup_slots)
    tup = tuple(dict.fromkeys(res).keys())
    if not (cls.__flags__ & 256):
        setattr(cls, '__slotnames__', list(tup))
    return tup



from dataclasses import dataclass
from typing import Any

import msgspec
from agentica_internal.core.json import JsonObject
from agentica_internal.warpc.kinds import (
    AnnotationsT,
    ClassT,
    FunctionT,
    ModuleT,
    ObjectT,
    TypeT,
    is_class_t,
    is_function_t,
    is_module_t,
    is_type_t,
)

type A = list[B]
type B = int | list[A]

json_schema = msgspec.json.schema


def ext_schema(obj: Any, root: bool = True) -> JsonObject:
    if is_type_t(obj):
        return annotation_schema(obj, root)
    if is_class_t(obj):
        return class_schema(obj, root)
    if is_function_t(obj):
        return function_schema(obj)
    elif is_module_t(obj):
        return module_schema(obj)
    else:
        return object_schema(obj)


def _schema_dict(kind: str, obj: object, **kwargs) -> JsonObject:
    return dict(type=kind, id=id(obj), **kwargs)


def function_schema(obj: FunctionT) -> JsonObject:
    annos, defs = json_schema_annos(getattr(obj, '__annotations__', {}))
    return _schema_dict('function', obj, annotations=annos, defs=defs)


def annotation_schema(anno: TypeT, root: bool = True) -> JsonObject:
    return _schema_dict('annotation', anno)
    # args = obj.__args__
    # name = f'{t_}Callable' if cls is TCallable else f'{a_}Callable'
    # if not args:
    #     return name
    # *dom, cod = args
    # if len(dom) == 1 and type(dom[0]) is EllipT:
    #     dom = dom[0]
    # pair = rec(dom), rec(cod)
    # return gen(name, pair)


def module_schema(mod: ModuleT) -> JsonObject:
    return _schema_dict('module', mod)


def object_schema(obj: ObjectT) -> JsonObject:
    cls = type(obj)
    cls_schema = class_schema(cls)
    cls_name = cls_schema['class_name']
    return _schema_dict('object', obj, class_name=cls_name, class_schema=cls_schema)


def class_schema(cls: ClassT, root: bool = True) -> JsonObject:
    schema = json_schema(foo)
    return _schema_dict(
        'class', cls, class_name=f'{cls.__module__}.{cls.__name__}', class_schema=schema
    )


def json_schema(obj) -> JsonObject:
    # check if msgspec WILL fail, if so dispatch to ext_schema with root = False
    ...
    try:
        return _json_schema(obj, schema_hook=json_schema_hook)
    except:
        return {'type': 'unrepresentable'}


def json_schema_annos(annos: AnnotationsT) -> tuple[JsonObject, JsonObject]:
    if not annos:
        return {}, {}
    try:
        schemas, defs = _json_schema_comps(annos.values(), schema_hook=json_schema_hook)
    except:
        defs = {}
        schemas = []
        for val in annos.values():
            val_schema, val_defs = _json_schema_comps((val,), schema_hook=json_schema_hook)
            defs.update(val_defs)
            schemas.append(val_schema)
    return dict(zip(annos.keys(), schemas)), defs


def class_name(cls: type) -> str:
    return f'{cls.__module__}.{cls.__name__}'


def json_schema_hook(obj: Any) -> dict:
    return {}


_json_schema = msgspec.json.schema
_json_schema_comps = msgspec.json.schema_components


@dataclass
class Clala:
    pass


def foo(i: int) -> Clala: ...


print(ext_schema(foo))



# fmt: off
from datetime import datetime
from collections.abc import Callable, Iterable as Iter
from re import MULTILINE, compile as compile_regex
from typing import Any

__all__ = [
    'ToStr',
    'Strs',
    'pascal_str',
    'camel_str',
    'upper_str',
    'snake_str',
    'indent_str',
    'unindent_str',
    'commas_str',
    'spaces_str',
    'lines_str',
    'indent_lines',
    'paren_str',
    'list_str',
    'tuple_str',
    'set_str',
    'dict_str',
    'dict_kv_str',
    'fn_call_str',
    'headed_str',
    'fn_call_str',
    'dict_str',
    'dict_kv_str',
    'pathsafe_str',
    'idsafe_str',
    'timestamp_str',
]

####################################################################################################

type ToStr = Callable[[object], str]
type Strs = Iter[str]

####################################################################################################

def pascal_str(snake: str) -> str:
    return ''.join(map(upper_str, snake.split('_')))


def camel_str(snake: str) -> str:
    if '_' not in snake:
        return snake
    first, *rest = snake.split('_')
    return first + ''.join(map(upper_str, rest))


def upper_str(word: str) -> str:
    if not word:
        return word
    return word[0].upper() + word[1:]


def snake_str(camel: str) -> str:
    s1 = UPPER_LOWER.sub(r'\1_\2', camel)
    return DIGIT_UPPER.sub(r'\1_\2', s1).lower()


def indent_str(s: str, i: int = 1) -> str:
    return s.replace('\n', indent_sep(i))


def unindent_str(s: str) -> str:
    return INDENTED_LINE.sub('', s)


def commas_str(seq: Strs) -> str:
    return ', '.join(seq)


def spaces_str(seq: Strs) -> str:
    return ' '.join(seq)


def lines_str(seq: Strs) -> str:
    return '\n'.join(seq)


def indent_sep(i: int) -> str:
    if i == 1:
        return '\n\t'
    if i == 0:
        return '\n'
    return '\n' + ('\t' * i)


def indent_lines(seq: Strs, i: int = 1) -> str:
    return indent_sep(i).join(seq)


def paren_str(seq: Strs) -> str:
    return f'({', '.join(seq)})'


def list_str(seq: Strs) -> str:
    seq = tuple(seq)
    if seq == ():
        return '[]'
    if has_nl(seq):
        return comma_indented_str('[', ']', seq)
    return f'[{', '.join(seq)}]'


def tuple_str(seq: Strs) -> str:
    seq = tuple(seq)
    if seq == ():
        return '(,)'
    if len(seq) == 1:
        return f'({seq[0]},)'
    if has_nl(seq):
        return comma_indented_str('(', ')', seq)
    return f'({', '.join(seq)})'


def set_str(seq: Strs) -> str:
    seq = tuple(seq)
    if len(seq) == 0:
        return 'set()'
    if has_nl(seq):
        return comma_indented_str('set(', ')', seq)
    return '{' + ', '.join(seq) + '}'


def frozenset_str(seq: Strs) -> str:
    seq = tuple(seq)
    if len(seq) == 0:
        return 'frozenset()'
    if has_nl(seq):
        return comma_indented_str('frozensetset(', ')', seq)
    return 'frozenset({' + ', '.join(seq) + '})'


def dict_str(dct: dict[str, Any], fn: ToStr, ml: bool = False, sort: bool = True) -> str:
    if not dct:
        return '{}'
    ks = []
    vs = []
    items = list(dct.items())
    if sort:
        items.sort(key=lambda pair: pair[0])
    for k, v in items:
        ks.append(k)
        vs.append(fn(v))
    return dict_kv_str(ks, vs, ml)


def dict_kv_str(keys: Strs, vals: Strs, ml: bool = False) -> str:
    if has_nl(vals) or ml:
        return comma_indented_str('{', '}', (f'{k!r}: {v}' for k, v in zip(keys, vals)))
    return '{' + ', '.join(f'{k!r}: {v}' for k, v in zip(keys, vals)) + '}'


def headed_str(head: str, seq: Strs) -> str:
    return f'{head}({','.join(seq)})'


def fn_call_str(fn: str, args: Strs, kwargs: dict[str, str]) -> str:
    args = list(args)
    args.extend(f'{k}={v}' for k, v in kwargs.items())
    f_args = ', '.join(args)
    return f'{fn}({f_args})'


def has_nl(args: Strs) -> bool:
    return any('\n' in arg for arg in args)


def comma_indented_str(l: str, r: str, args: Strs) -> str:
    return l + '\n  ' + ',\n  '.join(args) + '\n' + r


def pathsafe_str(text: str) -> str:
    text = text.replace('...', 'ellipsis')
    text = text.replace('[', '(').replace(']', ')').replace(', ', ',')
    safe = QUOTES.sub('', text)
    safe = NON_PATHSAFE.sub('_', safe)
    if len(safe) > 64:
        return safe[:64]
    return safe

def timestamp_str() -> str:
    # yymmdd-hhmm-micros
    # 251204-1428-340565
    return datetime.now().strftime("%y%m%d-%H%M%S-%f")

####################################################################################################

class IdReplacer(dict[..., str]):
    def __missing__(self, key) -> str:
        n = len(self)
        self[key] = rep = f'0x{n:08x}'
        return rep

def idsafe_str(s: str) -> str:
    if not ID_DIGITS.search(s):
        return s
    reps = IdReplacer()
    return ID_DIGITS.sub(lambda m: reps[m.group(0)], s)

####################################################################################################

UPPER_LOWER   = compile_regex('(.)([A-Z][a-z]+)')
DIGIT_UPPER   = compile_regex('([a-z0-9])([A-Z])')
INDENTED_LINE = compile_regex('^\t|    ', MULTILINE)
QUOTES        = compile_regex('[\'"]')
NON_ALNUM     = compile_regex(r'[^a-zA-Z0-9_]')
NON_PATHSAFE  = compile_regex(r'[^a-zA-Z0-9_()]')
ID_DIGITS     = compile_regex(r'\b[45678]\d{8,10}\b')



# fmt: off

from typing import TypeVar, Callable, Coroutine, overload
from pathlib import Path

from ...core.log import LoggingSpec
from ..events import Event
from .debug_world import DebugWorld


__all__ = [
    'ConnectedPair'
]

################################################################################

Type = TypeVar('Type', bound='type')
Return = TypeVar('Return', bound='object')
Object = TypeVar('Object', bound='object')


def foo(i: int, b: bool) -> str: pass

@overload
def sender(_: Type) -> Coroutine[None, None, Type]: ...
@overload
def sender(_: Callable[..., Return]) -> Callable[..., Coroutine[None, None, Return]]: ...
@overload
def sender(_: Object) -> Coroutine[None, None, Object]: ...


class ConnectedPair:
    a: DebugWorld
    b: DebugWorld

    def __init__(self,
                 a_name: str = 'a',
                 b_name: str = 'b',
                 logging: LoggingSpec = None,
                 dump_msgs: bool = False,
                 **kwargs) -> None:
        ...

    A = staticmethod(sender)
    B = staticmethod(sender)

    @property
    def tmp_file(self) -> Path: ...

    @property
    def worlds(self) -> tuple[DebugWorld, DebugWorld]: ...

    @property
    def pipes(self):
        return sender, sender

    async def __aenter__(self):
        return sender

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

    def history_str(self) -> str: ...

    def print(self) -> None: ...

    def collect_events(self, events: list[Event]): ...

    @property
    def last_a_event(self) -> Event: ...

    @property
    def last_b_event(self) -> Event: ...


ConnectedPair.A = staticmethod(sender)
ConnectedPair.B = staticmethod(sender)



# fmt: off

from collections.abc import Callable
from enum import Enum
from types import NotImplementedType
from typing import TYPE_CHECKING, Literal

__all__ = [
    'RequestHookFn',
    'PreRequestHookFn',
    'PostRequestHookFn',
    'RequestHook',
    'PreRequestHook',
    'PostRequestHook',
]

################################################################################

if TYPE_CHECKING:
    from ..core.result import Result
    from .resource.all import ResourceHandle, ResourceRequest
    from .resource.handle import ResourceHandle

################################################################################

type RequestHookFn = PreRequestHookFn | PostRequestHookFn
type PreRequestHookFn = Callable[[ResourceHandle, ResourceRequest], Result | NotImplementedType]
type PostRequestHookFn = Callable[[Result, ResourceHandle, ResourceRequest], Result | NotImplementedType]

################################################################################

class RequestHookType(Enum):
    PRE = 'pre'
    POST = 'post'


class RequestHook:
    hook: RequestHookFn
    when: RequestHookType

    def __init__(self, hook: RequestHookFn, when: RequestHookType):
        self.hook = hook
        self.when = when

    @classmethod
    def pre(cls, hook: RequestHookFn) -> 'PreRequestHook':
        return PreRequestHook(hook)

    @classmethod
    def post(cls, hook: RequestHookFn) -> 'PostRequestHook':
        return PostRequestHook(hook)


class PreRequestHook(RequestHook):
    hook: PreRequestHookFn
    when: Literal[RequestHookType.PRE] = RequestHookType.PRE

    def __init__(self, hook: RequestHookFn):
        self.hook = hook


class PostRequestHook(RequestHook):
    hook: PostRequestHookFn
    when: Literal[RequestHookType.POST] = RequestHookType.POST

    def __init__(self, hook: RequestHookFn):
        self.hook = hook



# fmt: off

from typing import TYPE_CHECKING

from ..alias import *

__all__ = [
    'ArgsMsg',
    'KwargsMsg',
    'AnnotationsMsg',
    'ResourcesRecordMsg',
    'AttributesMsg',
    'ClassesTupleMsg',
    'OverloadsMsg',
    'MethodMsg',
    'MethodsMsg',
]


################################################################################

if TYPE_CHECKING:
    from .resource_data import FunctionDataMsg
    from .term import TermMsg
    from .term_resource import ResourceMsg

################################################################################

# these make codec compilation easier

type ArgsMsg            = Tup['TermMsg']
type KwargsMsg          = Rec['TermMsg']
type AnnotationsMsg     = Rec['ResourceMsg']
type AttributesMsg      = Rec['TermMsg']
type ClassesTupleMsg    = Tup['ResourceMsg']
type OverloadsMsg       = Tup['FunctionDataMsg']
type ResourcesRecordMsg = Rec['ResourceMsg']
type MethodMsg          = tuple[MethodKind, 'ResourceMsg']
type MethodsMsg         = Rec[MethodMsg]



# ruff: noqa
# fmt: off

import pytest

from agentica_internal.testing.warpc.result_tests import verify_error_result, EXCEPTIONS


@pytest.mark.asyncio
@pytest.mark.parametrize('exception', EXCEPTIONS)
async def test_warp_type_annotation(exception):
    verify_error_result(exception)



# fmt: off

from enum import EnumType, Enum, IntEnum, StrEnum, ReprEnum, Flag, IntFlag

from .__ import *
from .base import *

__all__ = [
    'EnumType',
    'EnumKind',
    'EnumClassData',
]


################################################################################

type EnumKind = Literal['any', 'int', 'str', 'repr', 'flag', 'intflag']

class EnumClassData(ResourceData):
    __slots__ = 'kind', 'name', 'qualname', 'module', 'members', 'methods'

    FORBIDDEN_FORM = forbidden_class

    kind:     EnumKind
    name:     str
    qualname: optstr
    module:   optstr
    members:  AttributesT
    methods:  MethodsT

    # implementation attached later
    @classmethod
    def describe_resource(cls, enum_cls: EnumType) -> 'EnumClassData': ...

    # implementation attached later
    def create_resource(self, handle: ResourceHandle) -> EnumType: ...


################################################################################

_KINDS: list[EnumKind] = ['any', 'int', 'str', 'repr', 'flag', 'intflag']
_TYPES: list[EnumType] = [Enum, IntEnum, StrEnum, ReprEnum, Flag, IntFlag]

_TO_KIND: dict[EnumType, EnumKind] = dict(zip(_TYPES, _KINDS))
_TO_TYPE: dict[EnumKind, EnumType] = dict(zip(_KINDS, _TYPES))

################################################################################

def describe_real_enum_class(enum_cls: EnumType) -> EnumClassData:

    data = EnumClassData()
    data.kind = 'any'

    for base in enum_cls.__mro__:
        if kind := _TO_KIND.get(base):  # type: ignore
            data.kind = kind
            break

    name = enum_cls.__name__
    qualname = enum_cls.__qualname__
    module = enum_cls.__module__

    data.name = name
    data.qualname = qualname if qualname != name else None
    data.module = module
    data.methods = {}  # for now

    data.members = {k: v._value_ for k, v in enum_cls.__members__.items()}

    return data


################################################################################

def create_virtual_enum_class(data: EnumClassData, handle: ResourceHandle) -> ObjectT:

    handle.kind = Kind.Enum
    handle.keys = []
    handle.open = True
    handle.name = f'<{data.name!r} enum>'

    enum_type = _TO_TYPE[data.kind]

    v_cls = EnumType.__call__(
        enum_type, data.name,
        names=data.members, module=data.module, qualname=data.qualname
    )

    # TODO: attach methods

    return v_cls

################################################################################

EnumClassData.describe_resource = staticmethod(describe_real_enum_class)
EnumClassData.create_resource = create_virtual_enum_class



# fmt: off

from .alias import *

__all__ = [
    'UUIDs',
    'Ids',
    'MASK_16',
]


################################################################################

class Ids:
    __slots__ = 'i'

    i: int

    def __init__(self, i: int = 0):
        self.i = i

    def new(self) -> ID:
        self.i += 1
        return self.i

    __call__ = new


################################################################################

UUIDBase = Ids()
UUIDBase()

class UUIDs(Ids):
    __slots__ = (
        'm',
        't',
        'i',
    )

    m: int
    i: int

    def __init__(self):
        self.m = (time_ns_hash() & MASK_16) << 16
        super().__init__()

    def new(self) -> ID:
        self.i += 1
        return self.m ^ self.i

    __call__ = new

MASK_16 = 1 << 16 - 1


################################################################################

def local_name_hash() -> int:
    global _LOCAL_NAME_HASH
    if _LOCAL_NAME_HASH is not None:
        return _LOCAL_NAME_HASH
    import platform

    import xxhash

    name = platform.node()
    hash_int = xxhash.xxh64_intdigest(name)
    _LOCAL_NAME_HASH = hash_int
    return hash_int


_LOCAL_NAME_HASH: int | None = None


################################################################################

def time_ns_hash() -> int:
    import time

    import xxhash

    nanos = time.time_ns()
    hash_int = xxhash.xxh64_intdigest(nanos.to_bytes(32))
    return hash_int



# fmt: off

import asyncio
from inspect import iscoroutinefunction

from agentica_internal.core.type import anno_str
from agentica_internal.warpc.flags import with_flags
from agentica_internal.warpc.worlds.debug_world import DebugWorld, is_virtual


async def verify_future_function_ok():

    async def sqr(i: int) -> int:
        return i * 2

    assert iscoroutinefunction(sqr)

    with with_flags(DEFAULT_ASYNC_MODE='future'):

        pair = DebugWorld.connected_pair()
        async with pair:

            A, B = pair.pipes

            sqr_b = await B(sqr)
            assert is_virtual(sqr_b)

            assert not iscoroutinefunction(sqr_b)

            sqr_b_anno = anno_str(sqr_b.__annotations__)
            assert sqr_b_anno == "{'i': int, 'return': Future[int]}"

            sqr_a = await A(sqr_b)
            assert sqr_a is sqr

            sqr_b_future = sqr_b(5)
            assert isinstance(sqr_b_future, asyncio.Future)

            await asyncio.sleep(0.01)
            assert sqr_b_future.done()
            assert sqr_b_future.result() == 10

        return pair


async def verify_future_function_error():

    async def inv(i: int) -> float:
        return 1.0 / i

    assert iscoroutinefunction(inv)

    with with_flags(DEFAULT_ASYNC_MODE='future'):

        pair = DebugWorld.connected_pair()
        async with pair:

            A, B = pair.pipes

            inv_b = await B(inv)
            assert is_virtual(inv_b)

            assert not iscoroutinefunction(inv_b)

            inv_b_future = inv_b(0)
            assert isinstance(inv_b_future, asyncio.Future)

            await asyncio.sleep(0.01)
            assert inv_b_future.done()
            assert type(inv_b_future.exception()) is ZeroDivisionError

        return pair


if __name__ == '__main__':
    asyncio.run(verify_future_function_ok())
    asyncio.run(verify_future_function_error())



import msgspec
from agentica_internal.warpc.worlds.debug_world import DebugWorld

world = DebugWorld(name='world')
other_world = DebugWorld(name='other_world')


class ClsWithAttr:
    my_attr: str | None

    def __init__(self):
        self.my_attr = 'foo'


my_class = ClsWithAttr()

test_fiddlings: list[str | int] = ["one", 2, "three", 4, "five", 6, "seven", 8, "nine", 10]


def processIntersection(obj: ClsWithAttr) -> str:
    return "foo"


json_encoder = msgspec.json.Encoder()
msg = world.root.enc_any(processIntersection)
print([json_encoder.encode(ctx_def) for ctx_def in world.root.outgoing_defs])



# fmt: off

"""
This file exposes a bunch of internal class objects under more intelligible names.

# `types` module

From `types` we have simple aliases, like `FunctionT` -> `FunctionType`.

But more importantly, we have the following aliases for classes whose instances
are function-like / callable. The first column is the alias name, the second
is the name as reported by `types`, and the third

| alias                | original                  | example instance           |
| BoundMethodT         | MethodType                | `my_obj.my_inst_method`    |
| UnboundMethodC       | MethodDescriptorType      | `bytes.hex`                |
| UnboundDunderMethodC | WrapperDescriptorType     | `list.__init__`            |
| BoundDunderMethodC   | MethodWrapperType         | `[].__len__`               |
| BoundClassMethodC    | ClassMethodDescriptorType | `bytes.fromhex`            |
| BoundMethodOrFuncC   | BuiltinFunctionType       | `len`                      |
| BoundMethodOrFuncC   | BuiltinMethodType         | `[].append`                |
| MutablePropertyC     | GetSetDescriptorType      | `FrameType.f_locals`       |
| SlotPropertyC        | MemberDescriptorType      | `timedelta.days`           |

Note that `BuiltinFunctionType is BuiltinMethodType`, the instances of this
class have a `__self__` attribute that is a module for builtin functions,
e.g. `len.__self__ is builtins`, or a genuine object, e.g.
`lst.append.__self__ is lst`.

Here is what these various instances print as, if you want to tell at a glance
what you are dealing with from the print form of such an instance:
| BoundMethodT         | `<bound method 'NAME' of SELF>            |
| UnboundMethodC       | `<method 'NAME' of 'CLASS' objects>       |
| UnboundDunderMethodC | `<slot wrapper 'NAME' of 'CLASS' objects> |
| BoundDunderMethodC   | `<method-wrapper 'NAME' of SELF>          |
| BoundClassMethodC    | `<method 'NAME' of 'CLASS' objects>       |
| BoundMethodOrFuncC   | `<built-in function 'NAME'>`              |
|                      | `<built-in method 'NAME' of SELF>`        |
| MutablePropertyC     | `<attribute 'NAME' of 'CLASS' objects>    |
| SlotPropertyC        | `<slot 'NAME' of 'CLASS' objects>         |

Here is what the classes themselves print as:
| BoundMethodT         | `<class 'method'>'                     |
| UnboundMethodC       | `<class 'method_descriptor'>'          |
| UnboundDunderMethodC | `<class 'wrapper_descriptor'>'         |
| BoundDunderMethodC   | `<class 'method-wrapper'>'             |
| BoundClassMethodC    | `<class 'classmethod_descriptor'>'     |
| BoundMethodOrFuncC   | `<class 'builtin_function_or_method'>' |
| MutablePropertyC     | `<class 'getset_descriptor'>'          |
| SlotPropertyC        | `<class 'member_descriptor'>'          |

# `typing`, `collections.abc` module

To actually destructure type annotations, it helps to know what they are
underneath. E.g. `type(List[int])` is not a public class, and while all these
classes are quite rational, they have bizarre names, so here we expose them with
intelligible names.

The first letter of the alias `T` if it is from the `typing` module, and `C` if
it from the `types` module (since such classes are implemented in C).

The first column shows the alias, and the second column shows examples of this
class.

C-implemented in `types`:
| CGeneric       | `UserClass[int]` etc.
| CUnion         | `int | str`
| CCallable      | `collections.abc.Callable[[int], str]`

Python-implemented in `typing`:
| TBaseGeneric   | superclass, any of the below are instances
| TBlankGeneric  | `typing.Callable, typing.List, typing.Iterable` etc.
| TBlankCallable | ONLY `typing.Callable`
| TBlankTuple    | ONLY `typing.Tuple`
| TGeneric       | `typing.ClassVar[int], typing.TypeGuard[int]` etc.

Furthermore, `TGeneric` has various subclasses for special cases:
| TCallable      | `typing.Callable[[int], str]`
| TUnion         | `typing.Union[int, str]`
| TLiteral       | `typing.Literal['a', 'b']`
| TConcat        | `typing.Concat[...]`
| TUnpack        | `typing.Unpack[...]`
| TAnnotated     | `typing.Annotated[int, 'hello']`

Note that *unparameterized* forms of these type constructors are also valid
type annotations, e.g. `typing.Union` on its own. These are all instances
of the class `TForm`.

The following are in the own special category:
| TAny   | `typing.Any`
| TForm  | `typing.Never`, `typing.ClassVar`, `typing.Literal`, etc.

For completeness, we also expose the public classes:
| TForward       | ForwardRef     | `xxx: "whatever"`
| TAlias         | TypeAliasType  | `type xxx = whatever`
| TVar           | TypeVar        | the `X` in `def foo[X](): ...`
| TParamSpec     | ParamSpec      | the `X` in `def foo[**X](): ...`

## Utilities

To tell if an object is a type annotation, you can write:
`type(object) in ANNOS`.

To format a type annotation as a string, you can write:
`anno_str(anno)`, see its docstring for more info.
"""

from ..cpython.classes.sys import *
from ..cpython.classes.anno import *
from .anno import anno_str, is_anno  # noqa: F401



# fmt: off

import asyncio
import datetime as DT

from agentica_internal.warpc.worlds.debug_world import *


async def verify_plain_old_data():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair as B:

        async def assert_equal(a):
            b = await B(a)
            assert is_real(b), f"not is_real(b)"
            assert type(a) is type(b), f"{type(a)=}  is not  {type(b)=}"
            assert a is b, f"a is not b"

        async def assert_equivalent(a):
            b = await B(a)
            assert is_real(b), f"not is_real(b); {type(b)=}"
            assert type(a) is type(b), f"{type(a)=}  is not  {type(b)=}"
            assert a is not b, f"a is b"
            assert a == b, f"a != b"

        await assert_equal(True)
        await assert_equal(False)
        await assert_equal(None)

        await assert_equivalent(12345)
        await assert_equivalent(1.234)

        await assert_equal(())
        await assert_equivalent((1, 2, 3))

        await assert_equivalent([])
        await assert_equivalent([1, 2, 3])

        await assert_equivalent(set())
        await assert_equivalent({1, 2, 3})

        await assert_equal('')
        await assert_equivalent('string')

        await assert_equal(b'')
        await assert_equivalent(b'bytes')

        await assert_equivalent(DT.date(2023, 12, 25))
        await assert_equivalent(DT.time(14, 30, 45))
        await assert_equivalent(DT.timedelta(days=7, hours=3))
        await assert_equivalent(DT.datetime(2023, 12, 25, 14, 30, 45))

        await assert_equivalent((True, False, 5.3, [], 'string', b'bytes'))

    return pair


if __name__ == '__main__':
    asyncio.run(verify_plain_old_data())



# ruff: noqa
# fmt: off

from ..__ import *

from ..attrs import *
from ..forbidden import *
from ..kinds import *
from ..request.request_resource import *
from ..request.request_future import *
from ..msg.codec import EncoderP, DecoderP, CodecP

from .__raw import *

if TYPE_CHECKING:
    from ..msg.resource_data import *
    from ..msg.term_resource import *
    from .base import *
